from kaa.command import Commands, command, norec, norerun
from kaa import document
from kaa.ui.dialog import dialogmode
from kaa.theme import Theme, Style
from kaa.keyboard import *

MoveSeparatorThemes = {
    'basic': [],
}

moveseparator_keys = {
    left: 'moveseparator.prev',
    up: 'moveseparator.prev',
    (ctrl, 'p'): 'moveseparator.prev',
    (ctrl, 'b'): 'moveseparator.prev',
    right: 'moveseparator.next',
    down: 'moveseparator.next',
    (ctrl, 'f'): 'moveseparator.next',
    (ctrl, 'n'): 'moveseparator.next',
    '\r': 'moveseparator.close',
    '\n': 'moveseparator.close',
}


class MoveSeparatorCommands(Commands):

    @command('moveseparator.prev')
    @norec
    @norerun
    def prev(self, wnd):
        if wnd.document.mode.target:
            wnd.document.mode.target.separator_prev()

    @command('moveseparator.next')
    @norec
    @norerun
    def next(self, wnd):
        if wnd.document.mode.target:
            wnd.document.mode.target.separator_next()

    @command('moveseparator.close')
    @norec
    @norerun
    def close(self, wnd):
        # restore cursor
        org_wnd = wnd.document.mode.org_wnd
        org_wnd.activate()

        # Destroy popup window
        popup = wnd.get_label('popup')
        if popup:
            popup.destroy()

        org_wnd.activate()


class MoveSeparatorMode(dialogmode.DialogMode):
    USE_UNDO = False

    @classmethod
    def build(cls, target):
        buf = document.Buffer()
        doc = document.Document(buf)
        mode = cls()
        doc.setmode(mode)

        mode.org_wnd = target
        mode.target = target.splitter.parent

        with dialogmode.FormBuilder(doc) as f:
            f.append_text('caption',
                          'Hit cursor left/right key to resize window.')
        return doc

    def init_keybind(self):
        self.keybind.add_keybind(moveseparator_keys)

    def init_commands(self):
        super().init_commands()

        self.moveseparator_commands = MoveSeparatorCommands()
        self.register_command(self.moveseparator_commands)

    def init_themes(self):
        super().init_themes()
        self.themes.append(MoveSeparatorThemes)

    def is_cursor_visible(self):
        return 0   # hide cursor

    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        self.moveseparator_commands.close(wnd)

    def on_str(self, wnd, s):
        pass


def move_separator(wnd):
    doc = MoveSeparatorMode.build(wnd)
    kaa.app.show_dialog(doc)



from collections import namedtuple
from kaa.filetype.default import defaultmode
from gappedbuf import re as gre
from kaa.highlight import Tokenizer, Span, Keywords, EndSection, SingleToken
from kaa.theme import Theme, Style

JavaScriptThemes = {
    'basic': [],
}


def build_tokenizer(stop=None, terminates=None):
    JSTOKENS = namedtuple(
        'jstokens', ['keywords', 'number', 'comment1', 'comment2',
                     'string1', 'string2', 'stop'])
    keywords = Keywords('javascript-keyword', 'keyword',
                        ["break", "case", "catch", "continue", "debugger", "default", "delete",
                         "do", "else", "finally", "for", "function", "if", "in", "instanceof",
                         "new", "return", "switch", "this", "throw", "try", "typeof", "var",
                         "void", "while", "with", "class", "enum", "export", "extends", "import",
                         "super", "implements", "interface", "let", "package", "private",
                         "protected", "public", "static", "yield", ])

    number = SingleToken('javascript-numeric', 'number',
                         [r'\b[0-9]+(\.[0-9]*)*\b', r'\b\.[0-9]+\b'])
    comment1 = Span('javascript-comment1', 'comment',
                    r'/\*', r'\*/', escape='\\')
    comment2 = Span('javascript-comment2', 'comment', r'//', r'$', escape='\\')
    string1 = Span('javascript-string1', 'string', '"', '"', escape='\\')
    string2 = Span('javascript-string2', 'string', "'", "'", escape='\\')

    tokens = JSTOKENS(keywords, number, comment1,
                      comment2, string1, string2, stop)

    return Tokenizer(tokens, terminates=terminates)


class JavaScriptMode(defaultmode.DefaultMode):
    MODENAME = 'JavaScript'

    def init_themes(self):
        super().init_themes()
        self.themes.append(JavaScriptThemes)

    def init_tokenizers(self):
        self.tokenizers = [build_tokenizer()]



import io
import keyword
import copy
from kaa.filetype.default import defaultmode, theme
from kaa.highlight import Tokenizer, Keywords, Span, SingleToken
from kaa.theme import Theme, Style
from gappedbuf import re as gre
from kaa.command import Commands, command, norec, norerun
from kaa.keyboard import *
from kaa.ui.pythondebug import port

PythonThemes = {
    'basic': [
        Style('python-bytes', 'blue', 'default'),
    ],
}

PYTHONMENU = [
    ['&Comment', None, 'code.region.linecomment'],
    ['&Uncomment', None, 'code.region.unlinecomment'],
    ['&Table of contents', None, 'toc.showlist'],
    ['Toggle &Breakpoint', None, 'debugger.toggle.breakpoint'],
]

python_keys = {
    (ctrl, 't'): 'toc.showlist',
    f8: 'debugger.toggle.breakpoint',
}

KEYWORDS = ['and', 'as', 'assert', 'break', 'class', 'continue', 'def',
            'del', 'elif', 'else', 'except', 'finally', 'for', 'from',
            'global', 'if', 'import', 'in', 'is', 'lambda', 'nonlocal',
            'not', 'or', 'pass', 'raise', 'return', 'try', 'while',
            'with', 'yield']

CONSTANTS = ['False', 'None', 'True']


class PythonMode(defaultmode.DefaultMode):
    MODENAME = 'Python'
    re_begin_block = gre.compile(r"[^#]*:\s*(#.*)?$")
    LINE_COMMENT = '#'

    @classmethod
    def update_fileinfo(cls, fileinfo, document=None):
        import tokenize
        if not document:
            try:
                with open(fileinfo.fullpathname, 'rb') as buffer:
                    encoding, lines = tokenize.detect_encoding(buffer.readline)
                    fileinfo.encoding = encoding
            except IOError:
                pass
        else:
            s = document.gettext(0, 1024).encode('utf-8', errors='ignore')
            buffer = io.BytesIO(s)
            encoding, lines = tokenize.detect_encoding(buffer.readline)
            fileinfo.encoding = encoding

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, [python_keys])

    def init_menu(self):
        super().init_menu()
        self.menu['CODE'] = copy.deepcopy(PYTHONMENU)

    def init_themes(self):
        super().init_themes()
        self.themes.append(PythonThemes)

    def init_tokenizers(self):
        self.tokenizers = [Tokenizer([
            Keywords('python-statement', 'keyword', KEYWORDS),
            Keywords('python-constant', 'constant', CONSTANTS),
            SingleToken('python-numeric', 'number',
                        [r'\b[0-9]+(\.[0-9]*)*\b', r'\b\.[0-9]+\b']),
            Span('python-comment', 'comment', r'\#', '$', escape='\\'),
            Span('python-string31', 'string', 'r?"""', '"""', escape='\\'),
            Span('python-string32', 'string', "r?'''", "'''", escape='\\'),
            Span('python-string11', 'string', 'r?"', '"', escape='\\'),
            Span('python-string12', 'string', "r?'", "'", escape='\\'),

            Span('python-bytes31', 'python-bytes',
                 '(br?|r?b)"""', '"', escape='\\'),
            Span('python-bytes32', 'python-bytes',
                 "(br?|r?b)'''", "'''", escape='\\'),
            Span('python-bytes11', 'python-bytes',
                 '(br?|r?b)"', '"', escape='\\'),
            Span('python-bytes12', 'python-bytes',
                 "(br?|r?b)'", "'", escape='\\'),
        ])]

    def on_set_document(self, document):
        super().on_set_document(document)

        fname = document.get_filename()
        if fname:
            bps = port.get_breakpoints(fname)
            for bp in bps:
                pos = self.document.get_lineno_pos(bp.lineno)
                self.document.marks[bp] = pos

    def on_del_window(self, wnd):
        super().on_del_window(wnd)
        self.update_python_breakpoints()

    def get_line_overlays(self):
        ret = super().get_line_overlays()

        for k, v in self.get_breakpoints():
            ret[v] = 'breakpoint'
        return ret

    def on_file_saved(self, fileinfo):
        if self.document.get_filename() != fileinfo.fullpathname:
            self._clear_breakpoints()

        super().on_file_saved(fileinfo)

    RE_BEGIN_NEWBLOCK = gre.compile(r"[^#]*\:\s*(#.*)?$", gre.M)

    def on_auto_indent(self, wnd):
        pos = wnd.cursor.pos
        tol = self.document.gettol(pos)
        m = self.RE_BEGIN_NEWBLOCK.match(self.document.buf, tol, pos)
        if not m:
            super().on_auto_indent(wnd)
        else:
            f, t = self.get_indent_range(pos)
            t = min(t, pos)
            cols = self.calc_cols(f, t)
            indent = self.build_indent_str(cols + self.indent_width)
            indent = '\n' + indent
            self.insert_string(wnd, pos, indent, update_cursor=False)
            wnd.cursor.setpos(pos + len(indent))
            wnd.cursor.savecol()
            self._last_autoindent = (pos+1, wnd.cursor.pos)

    RE_SEARCH_NAME = gre.compile(r'(\\.)|(\w+)|(\S)')

    def _py_getname(self, pos):
        while True:
            m = self.RE_SEARCH_NAME.search(self.document.buf, pos)
            if not m:
                return self.document.endpos()
            if m.lastindex == 1:
                pos = m.end()
                continue
            elif m.lastindex == 2:
                return m.start(), m.end(), m.group()
            else:
                return m.start(), m.start(), ''

    def _py_def(self, m, token):
        pos, end, name = self._py_getname(m.end())
        if name:
            f, t = self.get_indent_range(m.start())
            cols = self.calc_cols(f, t)

            yield token, f, cols, name
        return end

    def _py_end_comment(self, m):
        close = r'(\\\n)|(\n)'
        reobj = gre.compile(close)

        pos = m.end()
        while True:
            m = reobj.search(self.document.buf, pos)
            if not m:
                return self.document.endpos()
            if m.lastindex == 2:
                return m.end()
            else:
                pos = m.end()

    def _py_end_string(self, m):
        close = r'\\.|(?P<CLOSE>%s)' % m.group()
        reobj = gre.compile(close)

        pos = m.end()
        while True:
            m = reobj.search(self.document.buf, pos)
            if not m:
                return self.document.endpos()
            if m.lastgroup == 'CLOSE':
                return m.end()
            else:
                pos = m.end()

    TOKENIZE_SEARCH_CLOSE = r'''
            (?P<BACKSLASH>\\.)|(?P<PARENTHESIS>[\{\[\(])|
            (?P<COMMENT>\#)|(?P<STRING>\"|\"\"\"|\'|\'\'\')'''

    def _py_close_parenthesis(self, m):
        close = '|(?P<CLOSE>\\%s)' % self.PARENSIS_PAIR[m.group()]
        reobj = gre.compile(self.TOKENIZE_SEARCH_CLOSE + close, gre.X)
        pos = m.end()
        while True:
            m = reobj.search(self.document.buf, pos)
            if not m:
                return self.document.endpos()

            g = m.lastgroup
            if g == 'CLOSE':
                return m.end()
            elif g == 'BACKSLASH':
                pos = m.end()
            elif g == 'PARENTHESIS':
                pos = self._py_close_parenthesis(m)
            elif g == 'STRING':
                pos = self._py_end_string(m)
            elif g == 'COMMENT':
                pos = self._py_end_comment(m)
            else:
                assert 0

    RE_TOKENIZE_SEARCH = gre.compile(
        r'''(?P<BACKSLASH>\\.)|(?P<CLASS>\bclass\b)|
            (?P<DEF>\bdef\b)|(?P<PARENTHESIS>[\{\[\(])|
            (?P<COMMENT>\#)|(?P<STRING>\"\"\"|\"|\'\'\'|\')''', gre.X)

    def _py_tokenize(self):
        pos = 0
        while True:
            m = self.RE_TOKENIZE_SEARCH.search(self.document.buf, pos)
            if not m:
                break
            g = m.lastgroup
            if g == 'BACKSLASH':
                pos = m.end()
            elif g == 'CLASS':
                pos = yield from self._py_def(m, 'class')
            elif g == 'DEF':
                pos = yield from self._py_def(m, 'def')
            elif g == 'PARENTHESIS':
                pos = self._py_close_parenthesis(m)
            elif g == 'STRING':
                pos = self._py_end_string(m)
            elif g == 'COMMENT':
                pos = self._py_end_comment(m)
            else:
                assert 0

    def get_headers(self):
        stack = []
        for token, pos, indent, name in self._py_tokenize():
            dispname = name if token is 'class' else name + '()'
            headertype = 'namespace' if token == 'class' else 'function'
            if not stack:
                header = self.HeaderInfo(
                    headertype, None, name, dispname, None, pos)
                yield header
                stack.append((indent, header))
                continue

            for i, (parent_indent, info) in enumerate(stack):
                if indent <= parent_indent:
                    del stack[i:]
                    break

            if stack:
                parents = tuple(header for (_, header) in stack
                                if header.token == 'namespace')
                parent = parents[-1] if parents else None
                fullname = '.'.join([stack[-1][1].name, name])
                header = self.HeaderInfo(
                    headertype, parent, fullname,
                    dispname, None, pos)
            else:
                header = self.HeaderInfo(
                    headertype, None, name, dispname, None, pos)

            yield header
            stack.append((indent, header))

    @command('toc.showlist')
    @norec
    @norerun
    def show_toclist(self, wnd):
        from kaa.ui.toclist import toclistmode
        toclistmode.show_toclist(wnd, list(self.get_headers()))

    def get_breakpoints(self):
        for k, v in self.document.marks.items():
            if isinstance(k, port.BreakPoint):
                yield k, v

    def update_python_breakpoints(self):
        filename = self.document.get_filename()
        if not filename:
            return

        bps = []
        for k, v in self.get_breakpoints():
            lineno = self.document.buf.lineno.lineno(v)
            k.lineno = lineno
            bps.append(k)

        port.set_breakpoints(filename, bps)

    def _clear_breakpoints(self):
        bps = [k for k in self.get_breakpoints()]
        for bp in bps:
            del self.document.marks[bp]

    @command('debugger.toggle.breakpoint')
    @norec
    @norerun
    def toggle_breakpoint(self, wnd):
        pos = wnd.cursor.pos
        tol = self.document.gettol(pos)
        for bp, pos in self.get_breakpoints():
            if tol == self.document.gettol(pos):
                del self.document.marks[bp]
                break
        else:
            filename = self.document.get_filename()
            if filename:
                lineno = self.document.buf.lineno.lineno(tol)
                bp = port.BreakPoint(filename, lineno)
                self.document.marks[bp] = tol

        self.document.style_updated()



import kaa
from kaa.filetype.default import modebase
from kaa.theme import Theme, Style


class StatusInfo:
    updated = False

    def __init__(self):
        self.infos = {
            'filename': '',
            'modified': '',
            'lineno': 1,
            'linecount': 1,
            'col': 1,
            'editmode': '',
        }

    def set_info(self, **values):
        ret = False
        for name, value in values.items():
            updated = True
            if name in self.infos:
                if value == self.infos[name]:
                    updated = False

            self.infos[name] = value

            self.updated = self.updated or updated
            ret = ret or updated
        return ret

    def get_info(self, name, default=None):
        return self.infos.get(name, default)


StatusBarThemes = {
    'basic': [
        Style('default', 'Red', 'Base02'),
        Style('filename', 'LightBlue', 'Base02'),
        Style('msg', 'Red', 'Base02'),
        Style('editmode', 'Green', 'Base02', rjust=True),
    ],
}


class StatusBarMode(modebase.ModeBase):
    USE_UNDO = False

    def __init__(self):
        super().__init__()
        self.statusinfo = StatusInfo()

    def init_themes(self):
        super().init_themes()
        self.themes.append(StatusBarThemes)

    def on_set_document(self, doc):
        super().on_set_document(doc)
        doc.undo = None

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

    def build_status(self, statusbar):
        d = statusbar.infos
        self.document.delete(0, self.document.endpos())
        style_default = self.get_styleid('default')
        style_filename = self.get_styleid('filename')
        style_editmode = self.get_styleid('editmode')

        if d['filename']:
            self.document.append(d['filename'], style_filename)

        if d['modified']:
            self.document.append('*', style_filename)

        self.document.append(' ', style_default)
        self.document.append(
            '[{lineno}:{col}] {linecount}'.format(**d), style_filename)

        self.document.append(d['editmode'], style_editmode)

    def set_info(self, **values):
        ret = self.statusinfo.set_info(**values)
        return ret

    def on_idle(self):
        if not self.closed:
            ret = super().on_idle()
            if not ret:
                if self.statusinfo.updated:
                    self.build_status(self.statusinfo)
                    self.statusinfo.updated = False
                    ret = True
            return ret



import kaa_testutils
from kaa import highlight
from kaa.filetype.css import cssmode


class TestHTMLHighlight(kaa_testutils._TestDocBase):
    tokenizers = [cssmode.build_tokenizer()]

    def test_comment(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc('/* ')

        assert [
            (0, 2, self.tokenizers[0].tokens[0].span_comment_begin),
            (2, 3, self.tokenizers[0].tokens[0].span_comment),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_selector(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc('/* */selector/**/selector')

        assert [
            (0, 2, self.tokenizers[0].tokens[0].span_comment_begin),
            (2, 3, self.tokenizers[0].tokens[0].span_comment),
            (3, 5, self.tokenizers[0].tokens[0].span_comment_end),
            (5, 13, self.tokenizers[0].tokens[0].span_selector),
            (13, 15, self.tokenizers[0].tokens[0].span_comment_begin),
            (15, 17, self.tokenizers[0].tokens[0].span_comment_end),
            (17, 25, self.tokenizers[0].tokens[0].span_selector),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_block(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc('{')

        assert [
            (0, 1, self.tokenizers[0].tokens[0].span_open_brace),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_block(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)
        doc = self._getdoc('{ abc:')

        assert [
            (0, 1, self.tokenizers[0].tokens[0].span_open_brace),
            (1, 5, self.tokenizers[0].tokens[0].span_propname),
            (5, 6, self.tokenizers[0].tokens[0].span_colon),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('{abc}')

        assert [
            (0, 1, self.tokenizers[0].tokens[0].span_open_brace),
            (1, 4, self.tokenizers[0].tokens[0].span_propname),
            (4, 5, self.tokenizers[0].tokens[0].span_close_brace),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('{/**/a/**/bc/**/}/**/')

        assert [
            (0, 1, self.tokenizers[0].tokens[0].span_open_brace),
            (1, 3, self.tokenizers[0].tokens[0].span_comment_begin),
            (3, 5, self.tokenizers[0].tokens[0].span_comment_end),
            (5, 6, self.tokenizers[0].tokens[0].span_propname),
            (6, 8, self.tokenizers[0].tokens[0].span_comment_begin),
            (8, 10, self.tokenizers[0].tokens[0].span_comment_end),
            (10, 12, self.tokenizers[0].tokens[0].span_propname),
            (12, 14, self.tokenizers[0].tokens[0].span_comment_begin),
            (14, 16, self.tokenizers[0].tokens[0].span_comment_end),
            (16, 17, self.tokenizers[0].tokens[0].span_close_brace),
            (17, 19, self.tokenizers[0].tokens[0].span_comment_begin),
            (19, 21, self.tokenizers[0].tokens[0].span_comment_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_prop(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('{abc:def')
        assert [
            (0, 1, self.tokenizers[0].tokens[0].span_open_brace),
            (1, 4, self.tokenizers[0].tokens[0].span_propname),
            (4, 5, self.tokenizers[0].tokens[0].span_colon),
            (5, 8, self.tokenizers[0].tokens[0].span_propvalue),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('{abc:def;ghi:jkl/**/;}')
        assert [
            (0, 1, self.tokenizers[0].tokens[0].span_open_brace),
            (1, 4, self.tokenizers[0].tokens[0].span_propname),
            (4, 5, self.tokenizers[0].tokens[0].span_colon),
            (5, 8, self.tokenizers[0].tokens[0].span_propvalue),
            (8, 9, self.tokenizers[0].tokens[0].span_semicolon),
            (9, 12, self.tokenizers[0].tokens[0].span_propname),
            (12, 13, self.tokenizers[0].tokens[0].span_colon),
            (13, 16, self.tokenizers[0].tokens[0].span_propvalue),
            (16, 18, self.tokenizers[0].tokens[0].span_comment_begin),
            (18, 20, self.tokenizers[0].tokens[0].span_comment_end),
            (20, 21, self.tokenizers[0].tokens[0].span_semicolon),
            (21, 22, self.tokenizers[0].tokens[0].span_close_brace),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_string(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('"aaa"')
        assert [
            (0, 1, self.tokenizers[0].tokens[0].span_string_begin),
            (1, 4, self.tokenizers[0].tokens[0].span_string),
            (4, 5, self.tokenizers[0].tokens[0].span_string_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc(r'"\""')
        assert [
            (0, 1, self.tokenizers[0].tokens[0].span_string_begin),
            (1, 3, self.tokenizers[0].tokens[0].span_string),
            (3, 4, self.tokenizers[0].tokens[0].span_string_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('{"aaa"}')
        assert [
            (0, 1, self.tokenizers[0].tokens[0].span_open_brace),
            (1, 2, self.tokenizers[0].tokens[0].span_string_begin),
            (2, 5, self.tokenizers[0].tokens[0].span_string),
            (5, 6, self.tokenizers[0].tokens[0].span_string_end),
            (6, 7, self.tokenizers[0].tokens[0].span_close_brace),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('{a:"aaa"}')
        assert [
            (0, 1, self.tokenizers[0].tokens[0].span_open_brace),
            (1, 2, self.tokenizers[0].tokens[0].span_propname),
            (2, 3, self.tokenizers[0].tokens[0].span_colon),
            (3, 4, self.tokenizers[0].tokens[0].span_string_begin),
            (4, 7, self.tokenizers[0].tokens[0].span_string),
            (7, 8, self.tokenizers[0].tokens[0].span_string_end),
            (8, 9, self.tokenizers[0].tokens[0].span_close_brace),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))



import kaa_testutils
from kaa.filetype.default import defaultmode, modebase


class TestDefaultMode(kaa_testutils._TestScreenBase):

    def test_getwords(self):
        mode = self._getmode()
        doc = self._getdoc('0123   \tabc ::::あいうえおカキクケコ[[')
        doc.setmode(mode)

        assert (
            (0, 4, '0123', 'L_WORDCHAR'),
            (4, 8, '   \t', 'Z_WHITESPACE'),
            (8, 11, 'abc', 'L_WORDCHAR'),
            (11, 12, ' ', 'Z_WHITESPACE'),
            (12, 16, '::::', 'Po'),
            (16, 21, 'あいうえお', 'L_HIRAGANA'),
            (21, 26, 'カキクケコ', 'L_KATAKANA'),
            (26, 28, '[[', 'Ps'),) == tuple(mode.split_word(0))

        mode = self._getmode()
        doc = self._getdoc('\n\n')
        doc.setmode(mode)

        assert (
            (0, 1, '\n', '_LF'),
        ) == tuple(mode.split_word(0))

    def test_search_next(self):
        w = self._getwnd('abcdefgabcdefg/efg/EFG')

        opt = modebase.SearchOption()
        opt.text = 'EFG'
        opt.ignorecase = True
        opt.word = False
        opt.regex = False

        ret = w.document.mode.search_next(0, opt)
        assert ret.span() == (4, 7)

        opt.word = True
        ret = w.document.mode.search_next(0, opt)
        assert ret.span() == (15, 18)

        opt.ignorecase = False
        ret = w.document.mode.search_next(0, opt)
        assert ret.span() == (19, 22)

        opt.text = r'[EFG]+'
        opt.regex = True
        ret = w.document.mode.search_next(0, opt)
        assert ret.span() == (19, 22)

        opt.text = r'[EFG]+'
        opt.regex = False
        ret = w.document.mode.search_next(0, opt)
        assert ret is None

    def test_search_prev(self):
        w = self._getwnd('abcdefgabcdefg/efg/EFG01234567890')

        opt = modebase.SearchOption()
        opt.text = 'EFG'
        opt.ignorecase = True
        opt.word = False
        opt.regex = False

        endpos = w.document.endpos()
        ret = w.document.mode.search_prev(endpos, opt).span()
        assert ret == (19, 22)

        opt.text = r'[EFG]+'
        opt.regex = False
        ret = w.document.mode.search_prev(endpos, opt)
        assert ret is None

    def test_iter_parenthsis(self):
        doc = self._getdoc("({[]})")
        doc.styles.replaceints(0, doc.endpos(), [1, 2, 3, 4, 5, 6])

        ret = [(pos, c, attr) for pos, c, attr in doc.mode.iter_parenthesis(0)]

        assert ret == [(0, '(', 1), (1, '{', 2), (2, '[', 3),
                       (3, ']', 4), (4, '}', 5), (5, ')', 6)]

    def test_iter_rev_parenthsis(self):
        doc = self._getdoc("({[]})")
        doc.styles.replaceints(0, doc.endpos(), [1, 2, 3, 4, 5, 6])

        ret = [(pos, c, attr)
               for pos, c, attr in doc.mode.iter_rev_parenthesis(5)]

        assert ret == [(5, ')', 6), (4, '}', 5), (3, ']', 4),
                       (2, '[', 3), (1, '{', 2), (0, '(', 1)]

    def test_find_pair_parenthesis(self):
        doc = self._getdoc("({[({[]})]})")
        doc.styles.replaceints(
            0, doc.endpos(), [1, 1, 1, 2, 2, 2, 2, 2, 2, 1, 1, 1])

        assert doc.mode.find_match_parenthesis(0) == 11



from collections import namedtuple
from kaa.filetype.default import defaultmode
from gappedbuf import re as gre
from kaa.highlight import Tokenizer, Span, SingleToken, Token, SubSection, EndSection
from kaa.theme import Theme, Style
from kaa.filetype.javascript import javascriptmode

CSSThemes = {
    'basic': [
        Style('css-selector', 'magenta', 'default'),
        Style('css-propname', 'blue', 'default', bold=True),
        Style('css-propvalue', 'green', 'default', bold=True),
    ],
}


class CSSProp(Token):

    def __init__(self, name, stylename, close='', closestyle=None):
        super().__init__(name, stylename)

        self.close = self.closestyle = ''
        if close:
            self.close = '(?P<CLOSE>{})|'.format(close)
            self.closestyle = closestyle

        self.CSS_TOP = gre.compile(
            self.close + r'(?P<COMMENT>/\*)|(?P<STRING>[\'"])|({)')
        self.END_COMMENT = gre.compile(r'\*/')
        self.PROPERTY_TOP = gre.compile(
            self.close + r'(?P<COMMENT>/\*)|(?P<STRING>[\'"])|:|}')
        self.PROPERTY_VALUE = gre.compile(
            self.close + r'(?P<COMMENT>/\*)|(?P<STRING>[\'"])|;|}')

    def re_start(self):
        return r'.'

    def prepare(self, tokenizer):
        super().prepare(tokenizer)

        self.span_cssws = self.assign_tokenid(tokenizer, 'default')

        self.span_comment_begin = self.assign_tokenid(tokenizer, 'comment')
        self.span_comment = self.assign_tokenid(tokenizer, 'comment')
        self.span_comment_end = self.assign_tokenid(tokenizer, 'comment')

        self.span_string_begin = self.assign_tokenid(tokenizer, 'string')
        self.span_string = self.assign_tokenid(tokenizer, 'string')
        self.span_string_end = self.assign_tokenid(tokenizer, 'string')

        self.span_selector = self.assign_tokenid(tokenizer, 'css-selector')
        self.span_open_brace = self.assign_tokenid(tokenizer, 'css-selector')
        self.span_close_brace = self.assign_tokenid(tokenizer, 'css-selector')

        self.span_propname = self.assign_tokenid(tokenizer, 'css-propname')
        self.span_colon = self.assign_tokenid(tokenizer, 'css-propname')

        self.span_propvalue = self.assign_tokenid(tokenizer, 'css-propvalue')
        self.span_colon = self.assign_tokenid(tokenizer, 'css-propvalue')
        self.span_semicolon = self.assign_tokenid(tokenizer, 'css-propvalue')
        if self.close:
            self.span_close = self.assign_tokenid(tokenizer, self.closestyle)

    def resume_pos(self, highlighter, tokenizer, doc, pos):
        # Returns top of current keyword
        if 0 < pos < len(doc.styles):
            # resume from top of last selector
            p = doc.styles.rfindint(
                [self.span_selector], 0, pos, comp_ne=False)
            if p == -1:
                return 0
            return p
        return 0

    def on_start(self, tokenizer, doc, pos, match):
        buf = doc.buf
        for f, t, tokenid in self.root_func(buf, match.start()):
            if pos != f:
                assert pos < f
                yield (pos, f, self.span_cssws)
            yield (f, t, tokenid)
            pos = t
            if self.close and tokenid == self.span_close:
                return pos, None, True
        return pos, None, False

    def parse_comment(self, buf, pos):
        m = self.END_COMMENT.search(buf, pos)
        if not m:
            if pos != len(buf):
                yield pos, len(buf), self.span_comment
            return len(buf)

        f, t = m.span()
        if pos != f:
            yield pos, f, self.span_comment

        if m.groupdict().get('CLOSE'):
            yield f, t, self.span_close
            return t

        yield f, t, self.span_comment_end
        return t

    def parse_string(self, buf, pos, qmark):
        reobj = gre.compile(r'(\\.)|({0})'.format(qmark), gre.DOTALL)
        cur = pos
        while True:
            m = reobj.search(buf, cur)
            if not m:
                yield pos, len(buf), self.span_string
                return len(buf)
            if m.group(1):
                cur = m.end()
            else:
                yield pos, m.start(), self.span_string
                yield m.start(), m.end(), self.span_string_end
                return m.end()

    def parse_properties(self, buf, pos):
        while True:
            m = self.PROPERTY_TOP.search(buf, pos)
            if not m:
                if pos != len(buf):
                    yield pos, len(buf), self.span_propname
                return len(buf)

            f, t = m.span()
            if pos != f:
                yield pos, f, self.span_propname

            if m.group('COMMENT'):
                yield f, t, self.span_comment_begin
                pos = yield from self.parse_comment(buf, t)

            elif m.group('STRING'):
                yield f, t, self.span_string_begin
                pos = yield from self.parse_string(buf, t, m.group('STRING'))

            elif m.group() == ':':
                yield f, t, self.span_colon
                pos = yield from self.parse_propvalue(buf, t)

            elif m.groupdict().get('CLOSE'):
                yield f, t, self.span_close
                return t

            else:
                yield f, t, self.span_close_brace
                return t

    root_func = parse_properties

    def parse_propvalue(self, buf, pos):
        while True:
            m = self.PROPERTY_VALUE.search(buf, pos)
            if not m:
                if pos != len(buf):
                    yield pos, len(buf), self.span_propvalue
                return len(buf)

            f, t = m.span()
            if pos != f:
                yield pos, f, self.span_propvalue

            if m.group('COMMENT'):
                yield f, t, self.span_comment_begin
                pos = yield from self.parse_comment(buf, t)

            elif m.group('STRING'):
                yield f, t, self.span_string_begin
                pos = yield from self.parse_string(buf, t, m.group('STRING'))

            elif m.group() == ';':
                yield f, t, self.span_semicolon
                return t

            elif m.groupdict().get('CLOSE'):
                yield f, t, self.span_close
                return t

            else:
                # close brace
                return f


class CSSToken(CSSProp):

    def parse_css(self, buf, pos):
        while True:
            m = self.CSS_TOP.search(buf, pos)
            if not m:
                if pos != len(buf):
                    yield pos, len(buf), self.span_selector
                return len(buf)

            f, t = m.span()
            if pos != f:
                yield pos, f, self.span_selector

            if m.group('COMMENT'):
                yield f, t, self.span_comment_begin
                pos = yield from self.parse_comment(buf, t)

            elif m.group('STRING'):
                yield f, t, self.span_string_begin
                pos = yield from self.parse_string(buf, t, m.group('STRING'))

            elif m.groupdict().get('CLOSE'):
                yield f, t, self.span_close
                return t

            else:
                yield f, t, self.span_open_brace
                pos = yield from self.parse_properties(buf, t)

    root_func = parse_css


def build_tokenizer(close=None, closestyle=None):
    csstoken = CSSToken('css', 'default', close=close, closestyle=closestyle)
    return Tokenizer([csstoken])


def build_proptokenizer(close=None, closestyle=None):
    csstoken = CSSProp('css', 'default', close=close, closestyle=closestyle)
    return Tokenizer([csstoken])


class CSSMode(defaultmode.DefaultMode):
    MODENAME = 'CSS'

    def init_themes(self):
        super().init_themes()
        self.themes.append(CSSThemes)

    def init_tokenizers(self):
        self.tokenizers = [build_tokenizer()]



import weakref
import re
import gappedbuf


def is_combine(c):
    if (('\u0300' <= c <= '\u036f') or  # Combining Diacritical Marks
            # Combining Diacritical Marks Supplement
            ('\u1dc0' <= c <= '\u1dff') or
            # Combining Diacritical Marks for Symbols
        ('\u20d0' <= c <= '\u20ff') or
            ('\ufe20' <= c <= '\ufe2f')):  # Combining Half Marks

        return True


class LineNo:

    def __init__(self):
        self.buf = gappedbuf.GappedBuffer()

    def linecount(self):
        return len(self.buf) + 1

    def lineno(self, pos):
        return self.bisect_left(pos) + 1

    def getpos(self, lineno):
        if lineno <= len(self.buf):
            return self.buf.getint(lineno - 1)

    def bisect_left(self, pos):
        lo = 0
        hi = len(self.buf)

        while lo < hi:
            mid = (lo + hi) // 2
            if self.buf.getint(mid) < pos:
                lo = mid + 1
            else:
                hi = mid
        return lo

    def inserted(self, pos, s):
        idx = 0
        lfs = []
        while True:
            idx = s.find('\n', idx)
            if idx == -1:
                break
            else:
                lfs.append(pos + idx)
            idx += 1

        lno = self.bisect_left(pos)
        self.buf.insertints(lno, lfs)

        d = len(s)
        for p in range(lno + len(lfs), len(self.buf)):
            self.buf.setints(p, p + 1, self.buf.getint(p) + d)

    def deleted(self, delfrom, delto):
        lno = self.bisect_left(delfrom)
        for idx in range(lno, len(self.buf)):
            pos = self.buf.getint(idx)
            if pos >= delto:
                break
        else:
            idx = len(self.buf)

        if lno != idx:
            self.buf.delete(lno, idx)

        d = delto - delfrom
        for p in range(lno, len(self.buf)):
            pos = self.buf.getint(p)
            d = min(delto, pos) - delfrom
            self.buf.setints(p, p + 1, pos - d)


class Buffer(gappedbuf.GappedBuffer):

    def __init__(self):
        self.listeners = []
        self.lineno = LineNo()

    def close(self):
        del self.listeners
        del self.lineno

    def insert(self, index, s):
        super().insert(index, s)
        self.lineno.inserted(index, s)
        self._updated(index, len(s), 0)

    def delete(self, begin, end):
        super().delete(begin, end)
        self.lineno.deleted(begin, end)
        self._updated(begin, 0, end - begin)

    def replace(self, begin, end, s):
        super().replace(begin, end, s)
        self.lineno.deleted(begin, end)
        self.lineno.inserted(begin, s)
        self._updated(begin, len(s), end - begin)

    def add_listener(self, listener):
        self.listeners.append(listener)

    def _updated(self, pos, inslen, dellen):
        for listener in self.listeners:
            listener(self, pos, inslen, dellen)


class Document:
    all = weakref.WeakSet()  # should not be used!
    temporary = False
    closed = False
    fileinfo = None

    @classmethod
    def find_filename(cls, filename):
        for doc in cls.all:
            if doc.get_filename() == filename:
                return doc

    def __init__(self, buf=None):
        self.wnds = []
        self.all.add(self)

        if buf is None:
            buf = Buffer()
        self.buf = buf
        self.buf.add_listener(self.updated)

        self.styles = gappedbuf.GappedBuffer()
        if len(self.buf):
            self.styles.insertints(0, [0] * len(self.buf))

        self.undo = Undo()
        self.marks = Marks()
        self.mode = None

        self.title = ''

    def add_window(self, wnd):
        self.wnds.append(wnd)
        self.mode.on_add_window(wnd)

    def del_window(self, wnd):
        self.mode.on_del_window(wnd)
        self.wnds.remove(wnd)
        if self.mode.CLOSE_ON_DEL_WINDOW:
            if not self.wnds:
                self.close()

    def close(self):
        """Close this document"""
        assert not self.wnds

        self.closed = True
        del self.wnds

        self.mode.close()
        self.buf.close()

        self.marks = self.buf = self.mode = None
        self.all.remove(self)

    def use_undo(self, is_useundo):
        if is_useundo:
            if not self.undo:
                self.undo = Undo()
        else:
            self.undo = None

    def set_title(self, title):
        self.title = title

    def get_title(self):
        title = '<untitled>'

        if self.title:
            title = self.title
        elif self.fileinfo and self.fileinfo.filename:
            title = self.fileinfo.filename

        return title

    def get_filename(self):
        filename = ''
        fileinfo = self.fileinfo
        if fileinfo:
            filename = fileinfo.fullpathname
        return filename

    def setmode(self, mode):
        if self.mode:
            self.mode.close()

        self.mode = mode
        mode.on_set_document(self)

    def updated(self, buf, pos, inslen, dellen):
        """Called when document updated"""

        if inslen:
            style = 0
            if 0 < pos <= len(self.styles):
                style = self.styles.getints(pos - 1, pos)[0]
            self.styles.replaceints(pos, pos + dellen, [style] * inslen)
        else:
            self.styles.delete(pos, pos + dellen)

        self.marks.updated(pos, inslen, dellen)
        self.update_screen(pos, inslen, dellen)

    def update_screen(self, pos, inslen, dellen):
        for wnd in self.wnds:
            wnd.on_document_updated(pos, inslen, dellen)

        if self.mode:
            self.mode.on_document_updated(pos, inslen, dellen)

    def style_updated(self, posfrom=0, posto=None):
        if posto is None:
            posto = self.endpos()

        for wnd in self.wnds:
            wnd.style_updated(posfrom, posto)

    def endpos(self):
        """Returns end position of this document(=size of document)."""
        return len(self.buf)

    def findchr(self, pos, chars):
        return self.buf.findchr(chars, pos, len(self.buf))

    def gettol(self, pos):
        """Returns top of line at pos"""

        if pos == 0:
            return 0
        tol = self.buf.rfindchr('\n', 0, pos)
        return 0 if tol == -1 else tol + 1

    def find_newline(self, pos):
        eol = self.buf.findchr('\n', pos, len(self.buf))
        if eol == -1:
            eol = len(self.buf)
        return eol

    def _findeol(self, pos):
        """Find next occurrence of newline"""

        eol = self.buf.findchr('\n', pos, len(self.buf))
        if eol != -1:
            eol += 1
        return eol

    def geteol(self, pos):
        """Returns top of next line or end of buffer"""

        eol = self._findeol(pos)
        if eol == -1:
            eol = len(self.buf)
        return eol

    def gettext(self, begin, end):
        return self.buf[begin:end]

    def getstyles(self, begin, end):
        return self.styles.getints(begin, end)

    def getline(self, pos):
        """Returns tuple of eol and string"""

        eol = self._findeol(pos)
        if eol == -1:
            return (len(self.buf), self.buf[pos:])
        else:
            return (eol, self.buf[pos:eol])

    def get_line_break(self, pos):
        eol = self.buf.findchr('\n', pos, len(self.buf))
        if eol == -1:
            return len(self.buf)
        return eol
        
    def iterlines(self, pos):
        end = self.endpos()
        while pos < end:
            eol = self.geteol(pos)
            yield self.buf[pos:eol]
            pos = eol

    def insert(self, pos, s, style=None):
        self.buf.insert(pos, s)
        if style is not None:
            self.styles.setints(pos, pos + len(s), style)

    def append(self, s, style=None):
        self.insert(self.endpos(), s, style)

    def delete(self, begin, end):
        self.buf.delete(begin, end)

    def replace(self, begin, end, s, style=None):
        self.buf.replace(begin, end, s)
        if style is not None:
            self.styles.setints(begin, begin + len(s), style)

    def get_nextpos(self, pos):
        pos += 1
        while pos < self.endpos():
            if not is_combine(self.buf[pos]):  # skip combine char
                return pos
            pos += 1
        return self.endpos()

    def get_prevpos(self, pos):
        while pos:
            pos -= 1
            if not is_combine(self.buf[pos]):  # skip combine char
                break
        return pos

    def get_lineno_pos(self, lineno):
        # line number starts from one.

        if lineno >= self.buf.lineno.linecount():
            return self.endpos()
        else:
            return self.buf.lineno.getpos(lineno)

    def setstyles(self, start, end, style, update=True):
        self.styles.setints(start, end, style)
        if update:
            self.style_updated(start, end)
            


class Marks(dict):
    locked = False

    def updated(self, pos, inslen, dellen):

        # don't update mark if locked
        if self.locked:
            return

        size = inslen - dellen
        if size == 0:
            return
        elif size > 0:
            for name, markpos in self.items():
                if isinstance(markpos, int):
                    if pos < markpos:
                        self[name] = markpos + size
                elif markpos is not None:
                    f, t = markpos
                    if pos < f:
                        f = f + size
                    if pos <= t:
                        t = t + size
                    self[name] = (f, t)

#            updated = ((name, markpos+size) for name, markpos in self.items()
#                            if (markpos is not None) and (pos < markpos))
        elif size < 0:
            size = abs(size)
            for name, markpos in self.items():
                if isinstance(markpos, int):
                    if pos < markpos:
                        if pos + size > markpos:
                            self[name] = pos
                        else:
                            self[name] = markpos - size
                elif markpos is not None:
                    f, t = markpos
                    if pos < f:
                        if pos + size > f:
                            f = pos
                        else:
                            f = f - size
                    if pos < t:
                        if pos + size > t:
                            t = pos
                        else:
                            t = t - size
                    self[name] = (f, t)


#            updated = ((name, pos if (pos+size > markpos) else (markpos-size))
#                            for name, markpos in self.items()
#                                if (markpos is not None) and (pos < markpos))
#        for name, p in updated:
#            self[name] = p
class Undo:

    """Records edit history"""

    def __init__(self):
        self.clear()

    def clear(self):
        self._actions = []      # edit operations
        self._closed = False    # Closed group undo block
        self._saved = 0         # position when document saved
        self._next_undo = 0     # undo action to be executed at next

    def _getblock(self):
        if self._actions:
            action, args, kwargs = self._actions[-1]
            if isinstance(action, Undo) and not action._closed:
                return action._getblock()
        return self

    def beginblock(self):
        """Begin group undo block"""

        block = self._getblock()
        block._add(Undo())

    def endblock(self):
        """End current group undo block"""

        block = self._getblock()
        assert block is not self
        if not block.can_undo():
            # Nothing happend
            del self._actions[-1]
            self._next_undo = len(self._actions)
        else:
            block._closed = True

    def _add(self, action, *args, **kwargs):
        # Can not redo thereafter
        del self._actions[self._next_undo:]
        if self._saved > len(self._actions):
            self._saved = -1

        self._actions.append((action, args, kwargs))
        self._next_undo = len(self._actions)

    def add(self, action, *args, **kwargs):
        """Add edit action"""

        self._getblock()._add(action, *args, **kwargs)

    def saved(self):
        """Notifies document saved"""

        self._saved = self._next_undo

    def is_dirty(self):
        """Returns True if buffer modified"""

        return self._next_undo != self._saved

    def can_undo(self):
        """Returns True if undo action is not exhausted"""

        return self._next_undo != 0

    def can_redo(self):
        """Returns True if redo action is not exhausted"""

        return self._next_undo < len(self._actions)

    def undo(self):
        """Perform undo action"""

        action, args, kwargs = self._actions[self._next_undo - 1]
        self._next_undo -= 1

        if isinstance(action, Undo):
            yield from action.undo_all()
        else:
            yield action, args, kwargs

    def undo_all(self):
        while self.can_undo():
            yield from self.undo()

    def redo(self):
        """Perform redo action"""

        action, args, kwargs = self._actions[self._next_undo]
        self._next_undo += 1
        if isinstance(action, Undo):
            yield from action.redo_all()
        else:
            yield action, args, kwargs

    def redo_all(self):
        while self.can_redo():
            yield from self.redo()



import sys
import subprocess
import shutil
import kaa

def select_clipboard():
    if sys.platform == 'darwin':
        return MacClipboard()
    elif X11Clipboard.check():
        return X11Clipboard()
    else:
        return Clipboard()

class Clipboard:
    MAX_CLIPBOARD = 10

    def _get_hist(self):
        return kaa.app.config.hist('clipboard', max_hist=self.MAX_CLIPBOARD)

    def get(self):
        all = self.get_all()
        ret = all[0] if all else ''
        if ret:
            self._get_hist().add(ret)
        return ret

    def get_all(self):
        return [s for s, i in self._get_hist().get() if s]

    def _set(self, s):
        self._get_hist().add(s)

    def set(self, s):
        self._set(s)
        
class NativeClipboard(Clipboard):
    def get(self):
        try:
            s = self._get_native_clipboard()
            if s:
                self._set(s)
            return s
        except Exception:
            kaa.log.error('Error to paste', exc_info=True)
            return super().get()

    def set(self, s):
        super().set(s)
        try:
            self._set_native_clipboard(s)
        except Exception as e:
            pass

class MacClipboard(NativeClipboard):
    COPYCOMMAND = 'pbcopy'
    PASTECOMMAND = 'pbpaste'
    def _get_native_clipboard(self):
        return subprocess.check_output(self.PASTECOMMAND, shell=True, 
                universal_newlines=True)

    def _set_native_clipboard(self, s):
        p = subprocess.Popen(self.COPYCOMMAND, stdin=subprocess.PIPE, 
                universal_newlines=True, shell=True)
        p.communicate(s)
        p.wait()

class X11Clipboard(NativeClipboard):
    CLIPCOMMAND = 'xsel'
    COPYCOMMAND = 'xsel -i -b'
    PASTECOMMAND = 'xsel -o -b'

    @classmethod
    def check(cls):
        if not shutil.which(X11Clipboard.CLIPCOMMAND):
            return False

        ret = subprocess.call(cls.CLIPCOMMAND, stdin=subprocess.PIPE, 
                universal_newlines=True, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE, shell=True)
        if ret:
            return False

        return True

    def _get_native_clipboard(self):
        return subprocess.check_output(self.PASTECOMMAND, shell=True, 
                universal_newlines=True)

    def _set_native_clipboard(self, s):
        p = subprocess.Popen(self.COPYCOMMAND, stdin=subprocess.PIPE, 
                universal_newlines=True, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE, shell=True)
        p.communicate(s)
        p.wait()



import subprocess
import collections
import socket
import json
import kaa
from kaa import document
from kaa.ui.msgbox import msgboxmode
from kaa.ui.inputline import inputlinemode

DEBUGGER = None


class InputReader:

    def get_reader(self):
        if DEBUGGER:
            return DEBUGGER.get_reader()
        else:
            return []

    def read_input(self, f):
        return DEBUGGER.read_input(f)

reader_initialized = False


def init_reader():
    global reader_initialized
    if reader_initialized:
        return
    reader_initialized = True
    # register input_reader
    kaa.app.add_input_reader(InputReader())


def _checkrunning():
    if DEBUGGER:
        if DEBUGGER.session:
            DEBUGGER.show_callstack()
            return True


def init_server():
    global DEBUGGER
    if _checkrunning():
        return

    if isinstance(DEBUGGER, RemoteDebugger):
        DEBUGGER.show_callstack()
        return
    else:
        if DEBUGGER:
            DEBUGGER.close()
            DEBUGGER = None

    def callback(w, s):
        s = s.strip()
        try:
            portno = int(s)
        except ValueError as e:
            kaa.app.messagebar.set_message(str(e))
            return

        global DEBUGGER
        DEBUGGER = RemoteDebugger()
        DEBUGGER.prepare(portno)

        init_reader()

    doc = inputlinemode.InputlineMode.build('Port no:',
                                            callback, value=str(
                                                RemoteDebugger.DEBUGGER_PORT),
                                            filter=inputlinemode.number_filter)

    kaa.app.show_dialog(doc)
    kaa.app.messagebar.set_message(
        'Enter debugger port number (default:28110)')


def run():
    global DEBUGGER
    if _checkrunning():
        return

    if DEBUGGER:
        DEBUGGER.close()
        DEBUGGER = None

    def callback(w, s):
        global DEBUGGER

        s = s.strip()
        if s:
            DEBUGGER = ChildDebugger()
            DEBUGGER.prepare()

            init_reader()

            kaa.app.config.hist('pythondebug_cmdline').add(s)
            DEBUGGER.run(s)

    hist = [s for s, info in kaa.app.config.hist('pythondebug_cmdline').get()]
    if hist:
        value = hist[0]
    else:
        value = 'python3.3 -m kaadbg.run myscript.py arg'

    doc = inputlinemode.InputlineMode.build('Command line:',
                                            callback, value=value, history=hist)
    kaa.app.messagebar.set_message(
        'Enter command line. (e.g python3.3 -m kaadbg.run myscript.py arg)')

    kaa.app.show_dialog(doc)


class BreakPoint:

    def __init__(self, filename, lineno):
        self.filename = filename
        self.lineno = lineno


breakpoints = []


def set_breakpoints(filename, bps):
    global breakpoints
    breakpoints = [b for b in breakpoints if b.filename != filename]
    breakpoints.extend(bps)


def get_breakpoints(filename):
    return [b for b in breakpoints if b.filename == filename]


def update_breakpoints():
    for doc in document.Document.all:
        if hasattr(doc.mode, 'update_python_breakpoints'):
            doc.mode.update_python_breakpoints()


class Debugger:
    port = None
    session = None
    current_frames = None
    debugpage = None
    running = False

    def _close_page(self):
        if self.debugpage and not self.debugpage.closed:
            self.debugpage.get_label('popup').destroy()
        self.debugpage = None

    def show_status(self):
        if self.debugpage and not self.debugpage.closed:
            if not self.port and self.is_idle():
                status = '-Closed-'
            elif self.is_idle():
                status = '-Accepting-'
            elif not self.running:
                status = '-BREAK-'
            else:
                status = '-RUNNING-'

            self.debugpage.document.mode.set_status(self.debugpage, status)
    
    def set_running(self, running):
        self.running = running

    def is_idle(self):
        return not self.session

    def close(self):
        global DEBUGGER
        DEBUGGER = None

        self.current_frames = None
        self._close_page()

        if self.session:
            self.session.close()
            self.session = None

        if self.port:
            self.port.close()
            self.port = None

    def get_reader(self):
        if not kaa.app.mainframe.is_idle():
            if kaa.app.focus is not self.debugpage:
                # Another form is active.
                return []

        debugger = self.session or self.port
        if debugger:
            return [debugger]
        else:
            return []

    def read_input(self, f):
        if f is self.port:
            self.accept()
        elif f is self.session:
            if not self.read_command():
                self.finished()

    def accept(self):
        if self.session:
            return

        try:
            conn, addr = self.port.accept()
        except BlockingIOError:
            return
        if addr[0] != '127.0.0.1':
            return
        conn.setblocking(True)
        self.session = conn
        self.current_frames = None
        self.breakpoints = []

    def readline(self):
        if not self.session:
            return ''

        L = []
        while True:
            c = self.session.recv(1)
            if not c:
                break
            L.append(c)
            if c == b'\n':
                break

        return b''.join(L)

    def show_callstack(self):
        if not self.debugpage or self.debugpage.closed:
            from kaa.ui.pythondebug import pythondebugmode
            wnd = pythondebugmode.show_callstack(self, self.current_frames)
            self.debugpage = wnd

        else:
            self.debugpage.document.mode.update(
                self.debugpage, self.current_frames)

        self.show_status()

    def show_expr_result(self, value):
        def callback(c):
            self.show_callstack()

        self.show_status()
        msgboxmode.MsgBoxMode.show_msgbox(
            'Value: {}'.format(value), ['&Ok'], callback, ['\r', '\n'])

    def read_command(self):
        datalen = self.readline()
        if not datalen:
            return

        datalen = int(datalen)
        if not datalen:
            return

        data = []
        while datalen:
            c = self.session.recv(datalen)
            if not c:
                return

            datalen -= len(c)
            data.append(c)

        self.set_running(False)

        type, value = json.loads(str(b''.join(data), 'utf-8'))
        if type == 'frame':
            self.current_frames = value
            self.show_callstack()
        elif type == 'expr':
            self.show_expr_result(value)
        return True

    def send(self, obj):
        s = (json.dumps(obj) + '\n').encode('utf-8')
        self.session.send(('%d\n' % len(s)).encode('utf-8'))
        self.session.send(s)

    def set_breakpoints(self):
        command = '''
self.clear_all_breaks()
'''
        for bp in breakpoints:
            command += '''
self.set_break({bp.filename!r}, {bp.lineno})
'''.format(bp=bp)
        self.send(('script', command))

    def set_step(self):
        self.set_breakpoints()
        self.show_status()

        command = '''
self.set_step()
ret.append(True)
'''
        self.send(('script', command))

    def set_next(self):
        self.set_breakpoints()
        self.set_running(True)
        self.show_status()

        command = '''
self.set_next(frame)
ret.append(True)
'''
        self.send(('script', command))

    def set_return(self):
        self.set_breakpoints()
        self.set_running(True)
        self.show_status()

        command = '''
self.set_return(frame)
ret.append(True)
'''
        self.send(('script', command))

    def set_continue(self):
        self.set_breakpoints()
        self.set_running(True)
        self.show_status()

        command = '''
self.set_continue()
ret.append(True)
'''
        self.send(('script', command))

    def set_quit(self):
        command = '''
self.set_quit()
ret.append(True)
'''
        self.send(('script', command))

    def get_breakpoints(self):
        return breakpoints

    def show_expr(self, depth, expr):
        command = '''
import io, traceback
try:
    for d in range({depth}):
        frame = frame.f_back
    ret = repr(eval({expr!r}, frame.f_globals, frame.f_locals))
    self.send((u'expr', ret))
except Exception:
    self.send((u'expr', get_tb()))
'''.format(expr=expr, depth=depth)

        self.set_running(True)
        self.show_status()
        self.send(('script', command))


class RemoteDebugger(Debugger):
    DEBUGGER_PORT = 28110

    def prepare(self, portno):

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(('', portno))
        s.listen(5)
        s.setblocking(False)
        self.port = s
        kaa.app.messagebar.set_message(
            'Python debugger server started on port %d.' % portno)

        RemoteDebugger.DEBUGGER_PORT = portno

        self.show_callstack()

    def finished(self):
        if self.session:
            self.session.close()
            self.session = None
            self.current_frames = None
            self._close_page()


class ChildDebugger(Debugger):

    def prepare(self):
        self.session, self.child = socket.socketpair(
            socket.AF_UNIX, socket.SOCK_STREAM)
        kaa.app.messagebar.set_message('Python debugger started.')

    def finished(self):
        self.close()

    def run(self, s):
        env = {
            'KAADBG_DOMAINPORT': str(self.child.fileno())
        }
        self.set_running(True)
        self.process = subprocess.Popen(s, shell=True,
                    pass_fds=(self.child.fileno(),),
                    env=env)
        self.child.close()
        kaa.app.messagebar.set_message("Process started.")

    def close(self):
        self.set_running(False)
        if self.session:
            try:
                self.set_quit()
            except BrokenPipeError:
                pass
        super().close()
        if self.process:
            self.process.wait()
        kaa.app.mainframe.refresh()




import copy
from collections import namedtuple
from kaa.filetype.default import defaultmode
from gappedbuf import re as gre
from kaa.highlight import Tokenizer, Token, Span, Keywords, EndSection, SingleToken
from kaa.theme import Theme, Style
from kaa.command import Commands, command, norec, norerun
from kaa.keyboard import *

MarkdownThemes = {
    'basic': [
        Style('header', 'Blue', None),
        Style('hr', 'Green', None),
        Style('strong', 'Magenta', None),
        Style('emphasis', 'Blue', None),
        Style('literal', 'Cyan', None),
        Style('reference', 'Red', None),
        Style('role', 'Cyan', None),
        Style('substitution', 'Green', None),
    ],
}


class LinkToken(Token):
    RE_END = gre.compile('^\S', gre.M + gre.X)

    def re_start(self):
        return r'\['

    def prepare(self, tokenizer):
        super().prepare(tokenizer)

        self.text = self.assign_tokenid(tokenizer, self.stylename)
        self.url = self.assign_tokenid(tokenizer, self.stylename)
        self.link = self.assign_tokenid(tokenizer, self.stylename)

    def get_close(self, doc, pos, chars):
        r = gre.compile(r'\\.|' + '|'.join('\\' + c for c in chars))
        while True:
            m = r.search(doc.buf, pos)
            if not m:
                return '', doc.endpos()

            h = m.group()
            if h[0] == '\\':
                pos = m.end()
            else:
                return h, m.end()

    def on_start(self, tokenizer, doc, pos, match):
        f = match.start()
        c, t = self.get_close(doc, match.end(), ']')
        if t != doc.endpos():
            next = doc.gettext(t, t + 1)
            if next == ':':
                yield (f, t + 1, self.text)
                return t + 1, None, False
            elif next == ' ':
                t += 1
        yield (f, t, self.text)

        m = gre.compile(r'\s*(\[|\()').match(doc.buf, t)
        if not m:
            return t, None, False

        c = m.group(1)
        if c == '[':
            c, end = self.get_close(doc, m.end(), ']')
            yield t, end, self.link
        else:
            c, end = self.get_close(doc, m.end(), '")')
            if c == '"':
                c, end = self.get_close(doc, end, '"')
                c, end = self.get_close(doc, end, ')')
            yield t, end, self.url

        return end, None, False


class ImageToken(LinkToken):

    def re_start(self):
        return r'!\['


class MDInline(Span):
    WS = ' \t\r\n'

    def on_start(self, tokenizer, doc, pos, match):
        if ((pos >= doc.endpos() - 1) or
                (doc.gettext(pos + 1, pos + 2) in self.WS)):

            yield pos, pos + 1, tokenizer.nulltoken
            return pos + 1, None, False

        ret = yield from super().on_start(tokenizer, doc, pos, match)
        return ret


HEADERS = r'=-'


def build_tokenizer():
    MARKDOWNTOKENS = namedtuple('markdowntokens',
                                ['escape', 'header1', 'header2',
                                 'hr', 'link', 'image',
                                 'strong1', 'strong2',
                                 'emphasis1', 'emphasis2', 'code1', 'code2'])

    return Tokenizer(MARKDOWNTOKENS(
        # escape
        SingleToken('md-escape', 'default', [r'\\.']),

        # header
        SingleToken('md-header1', 'header',
                    [r'^.+\n(?P<H1>[{}])(?P=H1)+$'.format(HEADERS)]),
        SingleToken('md-header2', 'header', [r'^\#{1,6}.*$']),

        # hr
        SingleToken('md-hr', 'hr', [r'^(\-{3,}|_{3,}|\*{3,})$']),

        # link
        LinkToken('md-link', 'reference'),

        # image
        ImageToken('md-image', 'reference'),

        # strong
        MDInline('md-strong1', 'strong', r'\*\*', r'\*\*|$', escape='\\'),
        MDInline('md-strong2', 'strong', r'__', r'__|$', escape='\\'),

        # emphasis
        MDInline('md-emphasis1', 'emphasis', r'\*', r'\*|$', escape='\\'),
        MDInline('md-emphasis2', 'emphasis', r'_', r'_|$', escape='\\'),

        # code
        MDInline('md-code1', 'literal', r'``', r'``', escape='\\'),
        MDInline('md-code2', 'literal', r'`', r'`', escape='\\'),
    ))


MDMENU = [
    ['&Table of contents', None, 'toc.showlist'],
]

md_keys = {
    (ctrl, 't'): 'toc.showlist',
}


class MarkdownMode(defaultmode.DefaultMode):
    MODENAME = 'Markdown'

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, [md_keys])

    def init_menu(self):
        super().init_menu()
        self.menu['CODE'] = copy.deepcopy(MDMENU)

    def init_themes(self):
        super().init_themes()
        self.themes.append(MarkdownThemes)

    def init_tokenizers(self):
        self.tokenizers = [build_tokenizer()]

    HEADER1 = r'^(?P<TITLE>.+)\n(?P<H1>[{}])(?P=H1)+$'.format(HEADERS)
    HEADER2 = r'^(?P<H2>\#{1,6})(?P<TITLE2>.+)$'

    RE_HEADER = gre.compile('|'.join([HEADER1, HEADER2]), gre.X | gre.M)

    def get_headers(self):
        stack = []
        pos = 0
        while True:
            m = self.RE_HEADER.search(self.document.buf, pos)
            if not m:
                break

            pos = m.end()
            name = m.group('TITLE') or m.group('TITLE2')
            name = name.strip()

            bar = m.group('H1')
            if bar:
                if bar == '=':
                    level = 0
                else:
                    level = 1
            else:
                level = len(m.group('H2')) - 1

            if not stack:
                header = self.HeaderInfo('namespace', None, name, name,
                                         None, m.start())
                yield header
                stack.append((level, header))
                continue

            for i, (parent_level, info) in enumerate(stack):
                if level <= parent_level:
                    del stack[i:]
                    break

            if stack:
                parent = stack[-1][1]
            else:
                parent = None

            header = self.HeaderInfo(
                'namespace', parent, name,
                name, None, m.start())
            yield header
            stack.append((level, header))

    @command('toc.showlist')
    @norec
    @norerun
    def show_toclist(self, wnd):
        from kaa.ui.toclist import toclistmode
        headers = list(self.get_headers())
        toclistmode.show_toclist(wnd, headers)



import re
import itertools
import unicodedata
import collections

import gappedbuf.re
import kaa
from kaa import keyboard, editmode
import kaa.log
from kaa import highlight
from kaa import theme
from kaa import screen


class SearchOption:
    RE = gappedbuf.re

    def __init__(self):
        self.text = ''
        self.replace_to = ''
        self.ignorecase = True
        self.word = False
        self.regex = False

    def clone(self):
        ret = self.__class__()
        ret.text = self.text
        ret.replace_to = self.replace_to
        ret.ignorecase = self.ignorecase
        ret.word = self.word
        ret.regex = self.regex
        return ret

    def get_regex(self):
        text = self.text
        if not self.regex:
            text = self.RE.escape(text)
        if self.word:
            text = r'\b' + text + r'\b'

        opt = self.RE.MULTILINE
        if self.ignorecase:
            opt += self.RE.IGNORECASE

        regex = self.RE.compile(text, opt)
        return regex

SearchOption.LAST_SEARCH = SearchOption()


class ModeBase:
    (UNDO_INSERT,
     UNDO_REPLACE,
     UNDO_DELETE) = range(3)

    CLOSE_ON_DEL_WINDOW = True

    SCREEN_NOWRAP = False
    NO_WRAPINDENT = False
    SCREEN_BUILD_ENTIRE_ROWS = False
    SHOW_LINENO = False
    USE_UNDO = True
    DOCUMENT_MODE = False
    HIGHLIGHT_CURSOR_ROW = False
    SHOW_CURSOR = True

    tab_width = 8
    indent_width = 4
    indent_tab = False
    auto_indent = True

    closed = False
    current_lastcommands = None
    theme = None
    highlight = None
    _check_fileupdate = 0
    _last_autoindent = None
    
    @classmethod
    def update_fileinfo(cls, fileinfo, document=None):
        pass

    def __init__(self):
        self.commands = {}
        self.is_availables = {}

        self.keybind = keyboard.KeyBind()
        self.keybind_vi_commandmode = keyboard.KeyBind()
        self.keybind_vi_visualmode = keyboard.KeyBind()
        self.keybind_vi_visuallinewisemode = keyboard.KeyBind()
        self.init_keybind()

        self.init_commands()
        self.init_menu()

        self.themes = []
        self.init_themes()
        self._build_theme()
        kaa.app.translate_theme(self.theme)

        self.tokenizers = []
        self.init_tokenizers()
        self.stylemap = {}
        self.stylenamemap = {}
        self.highlight = highlight.Highlighter(self.tokenizers)

        self.setup()

    def setup(self):
        pass

    def close(self):
        self.document = None
        self.closed = True

        self.keybind.clear()

        self.theme = None

        self.commands.clear()
        self.commands = None

        self.is_availables = None
        self.highlight.close()
        self.highlight = None
        self.tokenizers = None
        self.stylemap = None

    def _build_style_map(self):
        self.stylemap[0] = self.theme.get_style('default')
        for tokenid in self.highlight.tokenids:
            assert tokenid not in self.stylemap
            token = self.highlight.get_token(tokenid)
            if token:
                stylename = token.get_stylename(tokenid)
                style = self.theme.get_style(stylename)

                self.stylemap[tokenid] = style
                self.stylenamemap[style.name] = tokenid
            else:
                style = self.theme.get_style('default')
                self.stylemap[tokenid] = style
                self.stylenamemap[style.name] = tokenid

    def on_set_document(self, document):
        self.document = document
        self.document.styles.setints(0, len(self.document.styles), 0)

        self._build_style_map()
        self.document.use_undo(self.USE_UNDO)

    def on_document_updated(self, pos, inslen, dellen):
        self._last_autoindent = None
        if self.highlight:
            self.highlight.updated(self.document, pos, inslen, dellen)

    def on_file_saved(self, fileinfo):
        pass

    def on_add_window(self, wnd):
        self.editmode_insert(wnd)

    def on_del_window(self, wnd):
        if len(self.document.wnds) == 1:
            if self.DOCUMENT_MODE and self.document.fileinfo:
                self.document.fileinfo.save_screenloc(wnd.screen.pos)

    def on_focus(self, wnd):
        kaa.app.show_cursor(self.is_cursor_visible())

        # relocate cursor
        wnd.cursor.setpos(wnd.cursor.pos)

    def register_keys(self, keybind, keys):
        for d in keys:
            keybind.add_keybind(d)

    def init_keybind(self):
        pass

    def init_commands(self):
        for name in dir(self):
            attr = getattr(self, name)
            if hasattr(attr, 'COMMAND_ID') and callable(attr):
                self.commands[getattr(attr, 'COMMAND_ID')] = attr

    def init_menu(self):
        pass

    def init_themes(self):
        pass

    def init_tokenizers(self):
        pass

    def register_command(self, cmds):
        self.commands.update(cmds.get_commands())
        self.is_availables.update(cmds.get_commands_is_enable())

    def get_command(self, commandid):
        is_available = self.is_availables.get(commandid, None)
        cmd = self.commands.get(commandid, None)
        return (is_available, cmd)

    def editmode_insert(self, wnd):
        wnd.set_editmode(editmode.EditMode())

    def editmode_visual(self, wnd):
        wnd.set_editmode(editmode.VisualMode())

    def editmode_visual_linewise(self, wnd):
        wnd.set_editmode(editmode.VisualLinewiseMode())

    def editmode_command(self, wnd):
        wnd.set_editmode(editmode.CommandMode())

    def get_styleid(self, stylename):
        if stylename == 'default':
            return 0

        styleid = self.stylenamemap.get(stylename, None)
        if styleid is not None:
            return styleid
            
        if not self.stylemap:
            ret = 1
        else:
            ret = max(self.stylemap.keys()) + 1
            
        style = self.theme.get_style(stylename)
        self.stylemap[ret] = style
        self.stylenamemap[stylename] = ret
        return ret

    def select_theme(self, theme_name, themes):
        theme = themes.get(theme_name, None)
        if theme is None:
            theme = themes[kaa.app.DEFAULT_THEME]
        return theme

    def _build_theme(self):
        theme_name = kaa.app.get_current_theme()
        self.theme = theme.Theme([])
        for t in self.themes:
            self.theme.update(self.select_theme(theme_name, t))

    def get_style(self, tokenid):
        ret = self.stylemap.get(tokenid, None)
        if ret is None:
            ret = self.theme.get_style('default')
        return ret

    def get_menu(self, itemname):
        return self.menu.get(itemname, None)

    def is_cursor_visible(self):
        return self.SHOW_CURSOR

    def on_keypressed(self, wnd, event, s, commands, candidate):
        return s, commands, candidate

    def filter_string(self, wnd, s):
        return s

    def on_edited(self, wnd):
        pass

    def on_str(self, wnd, s):
        self.put_string(wnd, s)
        wnd.screen.selection.clear()

        if kaa.app.macro.is_recording():
            kaa.app.macro.record_string(s)
        if self.highlight:
            # run highlighter a bit to display changes immediately.
            self.highlight.update_style(self.document, batch=50)

    def add_lastcommands(self, commandids):
        if self.current_lastcommands is not None:
            self.current_lastcommands.extend(commandids)

    def on_commands(self, wnd, commandids):
        try:
            if callable(commandids):
                commandids(wnd)
                return

            self.current_lastcommands = []
            for commandid in commandids:
                is_available, command = self.get_command(commandid)
                if not command:
                    msg = 'command {!r} is not registered.'.format(commandid)
                    kaa.app.messagebar.set_message(msg)
                    kaa.log.error(msg)
                    return

                command(wnd)
                if kaa.app.macro.is_recording():
                    kaa.app.macro.record(command)

                if not getattr(command, 'NORERUN', False):
                    self.current_lastcommands.append(commandid)

            if self.current_lastcommands:
                kaa.app.lastcommands = self.current_lastcommands

        finally:
            wnd.editmode.clear_repeat()
            self.current_lastcommands = None

    def on_esc_pressed(self, wnd, event):
        pass

    def on_cursor_located(self, wnd, pos, y, x):
        pass

    def insert_string(self, wnd, pos, s, update_cursor=True):
        """Insert string"""

        cur_pos = wnd.cursor.pos

        wnd.document.insert(pos, s)

        if update_cursor:
            wnd.cursor.setpos(wnd.cursor.pos + len(s))
            wnd.cursor.savecol()

        if wnd.document.undo:
            wnd.document.undo.add(self.UNDO_INSERT, pos, s,
                                  cur_pos, wnd.cursor.pos)

        self.on_edited(wnd)

    def replace_string(self, wnd, pos, posto, s, update_cursor=True):
        """Replace string"""

        cur_pos = wnd.cursor.pos

        deled = wnd.document.gettext(pos, posto)
        wnd.document.replace(pos, posto, s)

        if update_cursor:
            wnd.cursor.setpos(pos + len(s))
            wnd.cursor.savecol()

        if wnd.document.undo:
            wnd.document.undo.add(self.UNDO_REPLACE, pos, posto, s,
                                  deled, cur_pos, wnd.cursor.pos)

        self.on_edited(wnd)

    def delete_string(self, wnd, pos, posto, update_cursor=True):
        """Delete string"""

        cur_pos = wnd.cursor.pos

        if pos < posto:
            deled = wnd.document.gettext(pos, posto)
            wnd.document.delete(pos, posto)

            if update_cursor:
                wnd.cursor.setpos(pos)
                wnd.cursor.savecol()

            if wnd.document.undo:
                wnd.document.undo.add(self.UNDO_DELETE, pos, posto, deled,
                                      cur_pos, wnd.cursor.pos)
            self.on_edited(wnd)


    def replace_rect(self, wnd, repto):
        if wnd.document.undo:
            wnd.document.undo.beginblock()
        try:
            (posfrom, posto, colfrom, colto
             ) = wnd.screen.selection.get_rect_range()

            for s in repto:
                if posto <= posfrom:
                    break

                sel = wnd.screen.selection.get_col_string(
                    posfrom, colfrom, colto)
                if sel:
                    f, t, org = sel
                    if org.endswith('\n'):
                        t = max(f, t - 1)
                    self.replace_string(wnd, f, t, s)
                    posto += (len(s) - (t - f))
                posfrom = wnd.document.geteol(posfrom)
        finally:
            if wnd.document.undo:
                wnd.document.undo.endblock()

    def put_string(self, wnd, s):
        s = wnd.document.mode.filter_string(wnd, s)

        if wnd.screen.selection.is_selected():
            if wnd.screen.selection.is_rectangular():
                if '\n' not in s:
                    self.replace_rect(wnd, itertools.repeat(s))
                else:
                    self.replace_rect(wnd, s.split('\n'))

            else:
                sel = wnd.screen.selection.get_selrange()
                f, t = sel
                self.replace_string(wnd, f, t, s)
        else:
            self.insert_string(wnd, wnd.cursor.pos, s)


    def update_charattr(self, wnd):
        if wnd.charattrs:
            wnd.charattrs.clear()
            wnd.screen.style_updated()
            return True

    def on_idle(self):
        if self.closed:
            return

        ret = self.run_highlight()
        return ret

    def get_line_overlays(self):
        return {}

    HIGHLIGHTBATCH = 300

    def run_highlight(self):
        if self.highlight:
            return self.highlight.update_style(
                self.document,
                batch=self.HIGHLIGHTBATCH)

    def _split_chars(self, begin, end):
        """split characters by character category."""

        s = self.document.gettext(begin, end)
        for key, chars in itertools.groupby(s, unicodedata.category):
            chars = ''.join(chars)
            end = begin + len(chars)
            yield begin, end, chars, key
            begin = end

    RE_WORDCHAR = r"(?P<L_WORDCHAR>[a-zA-Z0-9_]+)"
    RE_WHITESPACE = r"(?P<Z_WHITESPACE>[\t ]+)"
    RE_LF = r"(?P<_LF>\n)"
    RE_HIRAGANA = r"(?P<L_HIRAGANA>[\u3040-\u309f\u30fc]+)"
    RE_KATAKANA = r"(?P<L_KATAKANA>[\u30a0-\u30ff\u30fc]+)"

    RE_SPLITWORD = gappedbuf.re.compile('|'.join((
        RE_WORDCHAR, RE_WHITESPACE, RE_LF, RE_HIRAGANA, RE_KATAKANA)))

    def split_word(self, begin, all=False):
        """yield word in the document until line ends"""

        for m in self.RE_SPLITWORD.finditer(self.document.buf, pos=begin):
            # split unmatched characters by character category.
            f, t = m.span()
            if f != begin:
                yield from self._split_chars(begin, f)
            begin = t

            yield (f, t, m.group(), m.lastgroup)

            # finish if we reach '\n'
            if not all and m.lastgroup == '_LF':
                return

        yield from self._split_chars(begin, self.document.endpos())

    def get_word_at(self, pos):
        tol = self.document.gettol(pos)
        ret = None
        for f, t, s, cg in self.split_word(tol):
            if t <= pos:
                if cg != 'Z_WHITESPACE':
                    if t == pos and cg[0] == 'L':
                        return (f, t, cg)
                    else:
                        ret = (f, t, cg)
            elif cg == '_LF':
                break
            else:
                if cg == 'Z_WHITESPACE':
                    if f == pos:
                        return ret
                    else:
                        return
                return (f, t, cg)

        if ret and ret[1] == pos:
            return ret

    def get_word_list(self):
        words = collections.OrderedDict()
        for f, t, s, cg in self.split_word(0, all=True):
            if cg in {'Z_WHITESPACE', '_LF'}:
                continue
            if cg[0] in 'LMN':  # Letter, Mark, Number
                if len(s) > 2:
                    words[s] = None
        return list(words.keys())

    def search_next(self, pos, searchinfo):
        regex = searchinfo.get_regex()
        pos = min(max(0, pos), self.document.endpos())
        m = regex.search(self.document.buf, pos)
        return m

    def search_prev(self, pos, searchinfo):
        regex = searchinfo.get_regex()
        last = None
        for m in regex.finditer(self.document.buf, 0):
            span = m.span()
            if span[1] >= pos:
                break
            last = m
        return last

    def get_line_sel(self, wnd):
        doc = wnd.document
        sel = wnd.screen.selection.get_selrange()
        if not sel:
            f = t = wnd.cursor.pos
        else:
            f, t = sel

        tol = doc.gettol(f)

        t_tol = doc.gettol(t)
        if t != t_tol:
            eol = doc.geteol(t)
        else:
            eol = t

        return tol, eol

    def get_indent_range(self, pos):
        tol = self.document.gettol(pos)
        regex = gappedbuf.re.compile(self.RE_WHITESPACE)
        m = regex.match(self.document.buf, tol)
        if m:
            return m.span()
        else:
            return (tol, tol)

    def build_indent_str(self, col):
        if self.indent_tab:
            ctab = col // self.tab_width
            cspc = col % self.tab_width
            return '\t' * ctab + ' ' * cspc
        else:
            return ' ' * col

    def cancel_auto_indent(self, wnd):
        if self._last_autoindent:
            f, t = (min(max(0, p), self.document.endpos())
                      for p in self._last_autoindent)
            self._last_autoindent = None
            if f != t and wnd.cursor.pos == t:
                e = self.document.get_line_break(wnd.cursor.pos)
                if e == t:
                    s = self.document.gettext(f, t)
                    m = re.match(self.RE_WHITESPACE, s)
                    if m and m.group() == s:
                        self.delete_string(
                            wnd, f, t, update_cursor=False)
                        wnd.cursor.setpos(f)
                    
    def on_auto_indent(self, wnd):
        pos = wnd.cursor.pos
        f, t = self.get_indent_range(pos)
        indent = self.document.gettext(f, min(pos, t))
        if pos <= t:
            b = self.document.get_line_break(pos)
            self.insert_string(wnd, f, '\n', update_cursor=False)
            wnd.cursor.setpos(wnd.cursor.pos + 1)
            wnd.cursor.savecol()

            # if new line don't hava non-ws char,
            # thie line should be subject to cancel-indent.
            if t == b:
                self._last_autoindent = (f+1, wnd.cursor.pos)

        else:
            indent = '\n' + indent
            self.insert_string(wnd, pos, indent, update_cursor=False)
            wnd.cursor.setpos(pos + len(indent))
            wnd.cursor.savecol()
            self._last_autoindent = (pos+1, wnd.cursor.pos)
            
        wnd.cursor.savecol()

    def calc_cols(self, f, t):
        chars = self.document.gettext(f, t)
        (dispchrs, dispcols, positions,
         intervals) = screen.translate_chars(f, chars, self.tab_width,
                                             kaa.app.config.AMBIGUOUS_WIDTH)

        return sum(dispcols)

    def on_global_next(self, wnd):
        return

    def on_global_prev(self, wnd):
        return



from kaa.filetype.default import defaultmode
from kaa.highlight import Tokenizer, Span
from kaa.theme import Theme, Style

DiffThemes = {
    'basic': [
        Style('default', 'Base00', None),
        Style('fromfile', 'Orange', None),
        Style('tofile', 'Yellow', None),
        Style('hunk', 'Blue', None),
        Style('remove', 'Orange', None, bold=True),
        Style('add', 'Yellow', None, bold=True),
        Style('header', 'Orange', None),
    ],
}


def build_tokenizer():
    return Tokenizer([
        Span('diff-fromfile', 'fromfile', r'^---', '$'),
        Span('diff-tofile', 'tofile', r'^\+\+\+', '$'),
        Span('diff-hunk', 'hunk', r'^@@', '@@'),
        Span('diff-add', 'add', r'^\+', '$'),
        Span('diff-remove', 'remove', r'^-', '$'),
        Span('diff-header', 'header', r'^\S+', '$'),
    ])


class DiffMode(defaultmode.DefaultMode):
    MODENAME = 'Diff'

    def init_themes(self):
        super().init_themes()
        self.themes.append(DiffThemes)

    def init_tokenizers(self):
        self.tokenizers = [build_tokenizer()]



import kaa
from kaa import document, command
from kaa.ui.dialog import dialogmode
from kaa.theme import Theme, Style
from kaa.filetype.default import keybind
from kaa.commands import editorcommand
from kaa.keyboard import *

PasteLinesThemes = {
    'basic': [],
}

pastelines_keys = {
    (alt, '\r'): ('paste.lines'),
    (alt, '\n'): ('paste.lines'),
}


class PasteLinesMode(dialogmode.DialogMode):
    autoshrink = True
    USE_UNDO = True
    auto_indent = False

    KEY_BINDS = [
        keybind.cursor_keys,
        keybind.edit_command_keys,
        keybind.emacs_keys,
        keybind.macro_command_keys,
        pastelines_keys
    ]

    def init_themes(self):
        super().init_themes()
        self.themes.append(PasteLinesThemes)

    def close(self):
        super().close()
        self.target = None

    def init_keybind(self):
        super().init_keybind()

        self.register_keys(self.keybind, self.KEY_BINDS)

    def init_commands(self):
        super().init_commands()

        self.cursor_commands = editorcommand.CursorCommands()
        self.register_command(self.cursor_commands)

        self.edit_commands = editorcommand.EditCommands()
        self.register_command(self.edit_commands)

        self.screen_commands = editorcommand.ScreenCommands()
        self.register_command(self.screen_commands)

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

        wnd.set_cursor(dialogmode.DialogCursor(wnd,
                                               [dialogmode.MarkRange('pastetext')]))

        wnd.cursor.setpos(self.document.marks['pastetext'][0])

        kaa.app.messagebar.set_message("Paste text and hit alt+Enter")

    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message("Canceled")

    @command.command('paste.lines')
    @command.norec
    @command.norerun
    def paste_lines(self, w):
        f, t = self.document.marks['pastetext']
        s = self.document.gettext(f, t)

        self.target.document.mode.put_string(self.target, s)
        w.screen.selection.clear()

        kaa.app.messagebar.set_message("{} letters inserted".format(t - f))

        popup = w.get_label('popup')
        popup.destroy()

    @classmethod
    def build(cls, target):
        buf = document.Buffer()
        doc = document.Document(buf)
        mode = cls()
        doc.setmode(mode)
        mode.target = target

        with dialogmode.FormBuilder(doc) as f:
            # caption
            f.append_text('caption',
                          'Paste text here. Hit alt+Enter when finished:')
            f.append_text('default', '\n')
            f.append_text('default', '', mark_pair='pastetext')

        return doc



import kaa_testutils
from kaa import highlight
from kaa.filetype.markdown import markdownmode


class TestMarkdownHighlight(kaa_testutils._TestDocBase):
    tokenizers = [markdownmode.build_tokenizer()]
    tokens = tokenizers[0].tokens

    def test_header1(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('abc\n---')
        assert [
            (0, 7, self.tokens.header1.tokenid),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_header2(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('# abc')
        assert [
            (0, 5, self.tokens.header2.tokenid),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_hr(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('----')
        assert [
            (0, 4, self.tokens.hr.tokenid),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_link(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('[link][link]')
        assert [
            (0, 6, self.tokens.link.text),
            (6, 12, self.tokens.link.link),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('[link](url)')
        assert [
            (0, 6, self.tokens.link.text),
            (6, 11, self.tokens.link.url),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('[link](url "te)xt")')
        assert [
            (0, 6, self.tokens.link.text),
            (6, 19, self.tokens.link.url),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('[link]:')
        assert [
            (0, 7, self.tokens.link.text),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_image(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('![link][link]')
        assert [
            (0, 7, self.tokens.image.text),
            (7, 13, self.tokens.image.link),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('![link](url)')
        assert [
            (0, 7, self.tokens.image.text),
            (7, 12, self.tokens.image.url),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

    def test_emphasis(self):
        hl = highlight.Highlighter(tokenizers=self.tokenizers)

        doc = self._getdoc('**text**')
        assert [
            (0, 2, self.tokens.strong1.span_start),
            (2, 6, self.tokens.strong1.span_mid),
            (6, 8, self.tokens.strong1.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('__text__')
        assert [
            (0, 2, self.tokens.strong2.span_start),
            (2, 6, self.tokens.strong2.span_mid),
            (6, 8, self.tokens.strong2.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('*text*')
        assert [
            (0, 1, self.tokens.emphasis1.span_start),
            (1, 5, self.tokens.emphasis1.span_mid),
            (5, 6, self.tokens.emphasis1.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('_text_')
        assert [
            (0, 1, self.tokens.emphasis2.span_start),
            (1, 5, self.tokens.emphasis2.span_mid),
            (5, 6, self.tokens.emphasis2.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('``text``')
        assert [
            (0, 2, self.tokens.code1.span_start),
            (2, 6, self.tokens.code1.span_mid),
            (6, 8, self.tokens.code1.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('`text`')
        assert [
            (0, 1, self.tokens.code2.span_start),
            (1, 5, self.tokens.code2.span_mid),
            (5, 6, self.tokens.code2.span_end),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))

        doc = self._getdoc('` text`')
        assert [
            (0, 1, self.tokenizers[0].nulltoken),
            (1, 6, self.tokenizers[0].nulltoken),
            (6, 7, self.tokenizers[0].nulltoken),
        ] == list((f, t, style) for f, t, style in hl.highlight(doc, 0))



import re
import os
import kaa
from kaa import encodingdef, consts
from kaa.keyboard import *
from kaa.ui.dialog import dialogmode
from kaa.theme import Theme, Style
from kaa.filetype.default import modebase, keybind
from kaa.ui.msgbox import msgboxmode
from kaa.command import command, Commands, norec, norerun
from kaa.commands import editorcommand
from kaa.ui.selectlist import filterlist
from kaa.ui.selectfile import selectfile
from kaa.ui.itemlist import itemlistmode
from kaa.ui.grep import grepmode


class GrepOption(modebase.SearchOption):
    RE = re

    def __init__(self):
        super().__init__()
        self.tree = True
        self.directory = '.'
        self.filenames = '*.*'
        self.encoding = 'utf-8'
        self.newline = 'auto'

    def clone(self):
        ret = super().clone()
        ret.tree = self.tree
        ret.directory = self.directory
        ret.filenames = self.filenames
        ret.encoding = self.encoding
        ret.newline = self.newline
        return ret

GrepOption.LASTOPTION = GrepOption()


GrepDlgThemes = {
    'basic': [],
}


class GrepCommands(editorcommand.EditCommands):

    @command('grepdlg.field.next')
    @norec
    @norerun
    def field_next(self, wnd):
        searchfrom, searchto = wnd.document.marks['searchtext']
        dirfrom, dirto = wnd.document.marks['directory']
        filefrom, fileto = wnd.document.marks['filenames']

        if searchfrom <= wnd.cursor.pos <= searchto:
            wnd.cursor.setpos(dirfrom)
            wnd.screen.selection.set_range(dirfrom, dirto)

        elif dirfrom <= wnd.cursor.pos <= dirto:
            wnd.cursor.setpos(filefrom)
            wnd.screen.selection.set_range(filefrom, fileto)

        else:
            wnd.cursor.setpos(searchfrom)
            wnd.screen.selection.set_range(searchfrom, searchto)

    @command('grepdlg.history')
    @norec
    @norerun
    def grep_history(self, wnd):
        searchfrom, searchto = wnd.document.marks['searchtext']
        dirfrom, dirto = wnd.document.marks['directory']
        filefrom, fileto = wnd.document.marks['filenames']

        if searchfrom <= wnd.cursor.pos <= searchto:
            def callback(result):
                if result:
                    f, t = wnd.document.marks['searchtext']
                    wnd.document.replace(f, t, result)
                    wnd.cursor.setpos(f)
                    f, t = wnd.document.marks['searchtext']
                    wnd.screen.selection.set_range(f, t)

            filterlist.show_listdlg('Recent searches',
                                    [s for s,
                                        info in kaa.app.config.hist('grep_text').get(
                                        )],
                                    callback)

        elif dirfrom <= wnd.cursor.pos <= dirto:
            def callback(result):
                if result:
                    f, t = wnd.document.marks['directory']
                    wnd.document.replace(f, t, result)
                    wnd.cursor.setpos(f)
                    f, t = wnd.document.marks['directory']
                    wnd.screen.selection.set_range(f, t)

            hist = []
            for p, info in kaa.app.config.hist('grep_dirname').get():
                path = os.path.relpath(p)
                hist.append(path if len(path) < len(p) else p)

            filterlist.show_listdlg('Recent directories',
                                    hist, callback)

        else:
            def callback(result):
                if result:
                    f, t = wnd.document.marks['filenames']
                    wnd.document.replace(f, t, result)
                    wnd.cursor.setpos(f)
                    f, t = wnd.document.marks['filenames']
                    wnd.screen.selection.set_range(f, t)

            filterlist.show_listdlg('Recent filenames',
                                    [s for s,
                                        info in kaa.app.config.hist('grep_filename').get(
                                        )],
                                    callback)


grepdlg_keys = {
    '\r': ('grepdlg.field.next'),
    '\n': ('grepdlg.field.next'),
    up: ('grepdlg.history'),
    tab: ('grepdlg.select-dir'),
}


class GrepDlgMode(dialogmode.DialogMode):
    autoshrink = True

    KEY_BINDS = [
        keybind.cursor_keys,
        keybind.edit_command_keys,
        keybind.emacs_keys,
        keybind.macro_command_keys,
        grepdlg_keys,
    ]

    target = None

    def __init__(self, wnd=None):
        super().__init__()
        if isinstance(wnd.document.mode, grepmode.GrepMode):
            GrepOption.LASTOPTION = wnd.document.mode.grepoption
            self.target = wnd
        else:
            config = kaa.app.config

            if wnd and wnd.screen.selection.is_selected():
                f, t = wnd.screen.selection.get_selrange()
                s = wnd.document.gettext(f, t).split('\n')
                if s:
                    s = s[0].strip()
                    if s:
                        GrepOption.LASTOPTION.text = s

            grepdir = config.hist('grep_dirname').get()
            if grepdir:
                GrepOption.LASTOPTION.directory = grepdir[0][0]

            grepfiles = config.hist('grep_filename').get()
            if grepfiles:
                GrepOption.LASTOPTION.filenames = grepfiles[0][0]

        self.option = GrepOption.LASTOPTION

    def close(self):
        super().close()

    def init_keybind(self):
        super().init_keybind()

        self.register_keys(self.keybind, self.KEY_BINDS)

    def init_commands(self):
        super().init_commands()

        self.cursor_commands = editorcommand.CursorCommands()
        self.register_command(self.cursor_commands)

        self.edit_commands = editorcommand.EditCommands()
        self.register_command(self.edit_commands)

        self.screen_commands = editorcommand.ScreenCommands()
        self.register_command(self.screen_commands)

        self._grepcommands = GrepCommands()
        self.register_command(self._grepcommands)

    def init_themes(self):
        super().init_themes()
        self.themes.append(GrepDlgThemes)

    def on_add_window(self, wnd):
        super().on_add_window(wnd)

        cursor = dialogmode.DialogCursor(wnd,
                                         [dialogmode.MarkRange('searchtext'),
                                          dialogmode.MarkRange('directory'),
                                             dialogmode.MarkRange('filenames')])
        wnd.set_cursor(cursor)
        f, t = self.document.marks['searchtext']
        wnd.cursor.setpos(f)
        wnd.screen.selection.set_range(f, t)

    def build_document(self):
        with dialogmode.FormBuilder(self.document) as f:
            # search text
            f.append_text('caption', '   Search:')
            f.append_text('default', ' ')
            f.append_text('default', self.option.text, mark_pair='searchtext')
            f.append_text('default', '\n')

            # directory
            f.append_text('caption', 'Directory:')
            f.append_text('default', ' ')

            path = self.option.directory
            if path:
                p = os.path.relpath(path)
                path = p if len(p) < len(path) else path
            f.append_text('default', path, mark_pair='directory')
            f.append_text('default', '\n')

            # filename
            f.append_text('caption', 'Filenames:')
            f.append_text('default', ' ')
            f.append_text('default', self.option.filenames,
                          mark_pair='filenames')
            f.append_text('default', '\n')

            # working directory
            f.append_text('default', '(current dir)')
            f.append_text('default', ' ')
            f.append_text('default', os.getcwd())
            f.append_text('default', '\n')

            # buttons
            f.append_text('checkbox', '[&Search]',
                          shortcut_style='checkbox.shortcut',
                          on_shortcut=self.run_grep)

            f.append_text('checkbox', '[&Dir]',
                          mark_pair='select-dir',
                          shortcut_style='checkbox.shortcut',
                          on_shortcut=self._select_dir)

            f.append_text('checkbox', '[&Tree]',
                          mark_pair='search-tree',
                          on_shortcut=self.toggle_option_tree,
                          shortcut_style='checkbox.shortcut',
                          shortcut_mark='shortcut-t')

            f.append_text('checkbox', '[&Ignore case]',
                          mark_pair='ignore-case',
                          on_shortcut=self.toggle_option_ignorecase,

                          shortcut_style='checkbox.shortcut',
                          shortcut_mark='shortcut-i')

            f.append_text('checkbox', '[&Word]',
                          mark_pair='word',
                          on_shortcut=self.toggle_option_word,
                          shortcut_style='checkbox.shortcut',
                          shortcut_mark='shortcut-w')

            f.append_text('checkbox', '[&Regex]',
                          mark_pair='regex',
                          on_shortcut=self.toggle_option_regex,
                          shortcut_style='checkbox.shortcut',
                          shortcut_mark='shortcut-r')

            f.append_text(
                'checkbox', '[&Encoding:{}]'.format(self.option.encoding),
                mark_pair='enc',
                shortcut_style='checkbox.shortcut',
                on_shortcut=lambda wnd:
                wnd.document.mode.select_encoding(wnd))

            f.append_text(
                'checkbox', '[&Newline:{}]'.format(self.option.newline),
                mark_pair='newline',
                shortcut_style='checkbox.shortcut',
                on_shortcut=lambda wnd:
                wnd.document.mode.select_newline(wnd))

            self.update_option_style()
            kaa.app.messagebar.set_message(
                "Hit alt+S to begin search. Hit up to show history.")

    def _set_option_style(self, mark, style,
                          shortcutmark, shortcutstyle):
        f, t = self.document.marks[mark]
        self.document.setstyles(f, t, self.get_styleid(style))

        f = self.document.marks[shortcutmark]
        self.document.setstyles(f, f + 1, self.get_styleid(shortcutstyle))

    def _get_optionstylename(self, f):
        if f:
            return '.checked'
        else:
            return ''

    def update_option_style(self):
        style = self._get_optionstylename(self.option.tree)
        self._set_option_style('search-tree', 'checkbox' + style, 'shortcut-t',
                               'checkbox.shortcut' + style)

        style = self._get_optionstylename(self.option.ignorecase)
        self._set_option_style('ignore-case', 'checkbox' + style, 'shortcut-i',
                               'checkbox.shortcut' + style)

        style = self._get_optionstylename(self.option.word)
        self._set_option_style('word', 'checkbox' + style, 'shortcut-w',
                               'checkbox.shortcut' + style)

        style = self._get_optionstylename(self.option.regex)
        self._set_option_style('regex', 'checkbox' + style, 'shortcut-r',
                               'checkbox.shortcut' + style)

    def _option_updated(self):
        self.update_option_style()
        self.document.style_updated()

    def _select_dir(self, wnd):
        def cb(dir):
            wnd.set_visible(True)
            if dir:
                path = os.path.relpath(dir)
                self.set_dir(wnd, path)

        wnd.set_visible(False)
        dir = os.path.abspath(self.get_dir())
        selectfile.show_selectdir(dir, cb)

    @command('grepdlg.select-dir')
    @norec
    @norerun
    def select_dir(self, wnd):
        f, t = self.document.marks['directory']
        if f <= wnd.cursor.pos <= t:
            self._select_dir(wnd)

    def toggle_option_tree(self, wnd):
        self.option.tree = not self.option.tree
        self._option_updated()

    def toggle_option_ignorecase(self, wnd):
        self.option.ignorecase = not self.option.ignorecase
        self._option_updated()

    def toggle_option_word(self, wnd):
        self.option.word = not self.option.word
        self._option_updated()

    def toggle_option_regex(self, wnd):
        self.option.regex = not self.option.regex
        self._option_updated()

    def _get_encnames(self):
        return sorted(encodingdef.encodings + ['japanese'],
                      key=lambda k: k.upper())

    def select_encoding(self, wnd):
        encnames = self._get_encnames()

        def callback(n):
            if n is None:
                return

            enc = encnames[n]
            if enc != self.option.encoding:
                self.option.encoding = enc
                f, t = self.document.marks['enc']
                # [Encoding:{mode}]
                # 01234567890    10
                self.document.replace(f + 10, t - 1, self.option.encoding)

        doc = itemlistmode.ItemListMode.build(
            'Select character encoding:',
            encnames,
            encnames.index(self.option.encoding),
            callback)

        kaa.app.show_dialog(doc)

    def select_newline(self, wnd):
        def callback(n):
            if n is None:
                return

            nl = consts.NEWLINES[n]
            if nl != self.option.newline:
                self.option.newline = nl
                f, t = self.document.marks['newline']
                # [Newline:{mode}]
                # 0123456789    10
                self.document.replace(f + 9, t - 1, self.option.newline)

        doc = itemlistmode.ItemListMode.build(
            'Select newline mode:',
            consts.NEWLINES,
            consts.NEWLINES.index(self.option.newline),
            callback)

        kaa.app.show_dialog(doc)

    def get_search_str(self):
        f, t = self.document.marks['searchtext']
        return self.document.gettext(f, t)

    def get_dir(self):
        f, t = self.document.marks['directory']
        return self.document.gettext(f, t)

    def set_dir(self, wnd, dir):
        f, t = self.document.marks['directory']
        self.replace_string(
            wnd, f, t, dir, update_cursor=True)

    def get_files(self):
        f, t = self.document.marks['filenames']
        return self.document.gettext(f, t)

    def run_grep(self, wnd):
        self.option.text = self.get_search_str()
        self.option.directory = self.get_dir()
        self.option.filenames = self.get_files()

        if (self.option.text and self.option.directory and
                self.option.filenames):

            kaa.app.config.hist('grep_text').add(self.option.text)
            path = os.path.abspath(self.option.directory)
            kaa.app.config.hist('grep_dirname').add(path)
            kaa.app.config.hist('grep_filename').add(self.option.filenames)

            grepmode.grep(self.option, self.target)

        wnd.get_label('popup').destroy()

    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        wnd.get_label('popup').destroy()
        kaa.app.messagebar.set_message("")



import curses
import curses.panel
import collections

from kaa import log
from kaa.cui import keydef
import kaa.context
from kaa import editmode


class Window(kaa.context.Context):

    """Base class to wrapper of curses.window object

    Class attributes:
        mainframe -- Mainframe of the window
    """

    mainframe = None

    WAITINPUTFOR = 1000  # wait 1000ms for input
    closed = False
    editmode = None

    def __init__(self, parent, wnd=None, pos=None):
        """Wrap window object wnd. Create new window if wnd was omitted."""

        super().__init__()

        w, h = 1, 1
        x = y = 0
        if parent:
            w, h = parent.getsize()

        if wnd:
            self._cwnd = wnd
        else:
            if pos:
                x, y = pos
            else:
                x = y = 0

            w = max(1, w)
            h = max(1, h)
            self._cwnd = curses.newwin(h, w, y, x)

        self._panel = curses.panel.new_panel(self._cwnd)
        self._cwnd.keypad(1)
        self.parent = parent
        self.children = []
        if self.parent:
            self.parent.children.append(self)

        self.rect = (x, y, x + w, y + h)

        self._oninit()

    def _oninit(self):
        pass

    def walk_children(self):
        # breadth first
        stack = collections.deque([self])
        while stack:
            v = stack.pop()
            if v is not self:
                yield v
            stack.extendleft(v.children)

    def destroy(self):
        """Destroy window"""
        if self.closed:
            return
        self.closed = True
        while self.children:
            self.children[0].destroy()

        if self.parent:
            self.parent.children.remove(self)
        self.parent = None
        self._cwnd = self._panel = None

        if kaa.app.focus is self:
            kaa.app.focus = None

    def get_context_parent(self):
        return self.parent

    _lasterror = ''
    _skipped = 0

    def do_input(self, nonblocking):
        """Get an input from curses"""
        self.bring_top()
        if nonblocking:
            self._cwnd.timeout(0)
        else:
            self._cwnd.timeout(-1)
#            self._cwnd.timeout(self.WAITINPUTFOR)

        no_trailing_char = False  # true if esc key pressed
        try:
            curses.panel.update_panels()
            curses.doupdate()

            c = self._cwnd.get_wch()
            if c == '\x1b':
                self._cwnd.timeout(0)
                try:
                    c2 = self._cwnd.get_wch()
                    # has trailing character.
                    curses.unget_wch(c2)
                except curses.error:
                    # no trailing character.
                    no_trailing_char = True

        except curses.error as e:
            if repr(e.args) != self._lasterror:
                if e.args[0] != 'no input':
                    log.debug('Error in get_wch()', exc_info=True)
                self._lasterror = repr(e.args)
            else:
                self._skipped += 1
                if self._skipped % 500 == 0:
                    log.debug(
                        'Too much input-error skips!: {} times'.format(self._skipped))
            return []

        self._lasterror = ''
        self._skipped = 0

        if c == curses.KEY_MOUSE:
            mouseid, x, y, z, bstate = curses.getmouse()
            keys = [keydef.MouseEvent(self, mouseid, x, y, z, bstate)]
        else:
            keys = [keydef.KeyEvent(self, k, no_trailing_char)
                    for k in keydef.convert_registered_key(c)]

#        for k in keys:
#            _trace('{!r}'.format(k))

        return keys

    def add_str(self, letters, attr):
        try:
            self._cwnd.addstr(letters, attr)
        except curses.error:
            # exception is raised if a characters are written
            # over right border.
#            logstr = letters if (len(letters) < 10) else (letters[:10]+'...')
#            log.debug('Error to write str:{!r}'.format(logstr), exc_info=True)
            pass

    def bring_top(self):
        self._panel.top()

    def activate(self):
        """Activate wnd"""
        self.bring_top()
        kaa.app.set_focus(self)

    def set_rect(self, l, t, r, b):
        """Set window coordinate"""

        if hasattr(self, 'splitter') and self.splitter:
            assert t <= self.splitter.rect[1]

        try:
            self._cwnd = curses.newwin(b - t, r - l, t, l)
            self.rect = (l, t, r, b)
        except curses.error as e:
            log.debug('error on resizing window: {} {}'.format(self,
                                                               (l, t, r, b), exc_info=True))
            # shrink window to avoid segfault in curses
            self._cwnd = curses.newwin(1, 1, 0, 0)
            self.rect = (0, 0, 1, 1)

         # surprisingly enough, the following assert fails sometime.
#        t, l = self._cwnd.getbegyx()
#        h, w = self._cwnd.getmaxyx()
#        assert (l, t, l+w, t+h) == self.rect

        self._cwnd.keypad(1)

        self._panel = curses.panel.new_panel(self._cwnd)
        self.on_setrect(*self.rect)

    def draw_screen(self, force=False):
        pass

    def getsize(self):
        h, w = self._cwnd.getmaxyx()
        return w, h

    def refresh(self):
        self.mainframe.request_refresh()

    def update_window(self):
        return self.draw_screen()

    def update_status(self):
        return False

    def on_idle(self):
        return

    def on_setrect(self, l, t, r, b):
        """Window resized"""
        pass

    def on_focus(self):
        """Got focus"""
        pass

    def on_killfocus(self):
        pass



from kaa.filetype import filetypedef

class FileTypeInfo(filetypedef.FileTypeInfo):
    FILE_EXT = {'.htm', '.html'}

    @classmethod
    def get_modetype(cls):
        from kaa.filetype.html.htmlmode import HTMLMode
        return HTMLMode




import kaa
import string


class EditMode:
    MODENAME = 'Insert'

    pending_str = ''
    repeat = None
    repeat_str = ''

    def __init__(self):
        self.pending_keys = []
        self.last_command_keys = []

    def activated(self, wnd):
        pass

    def flush_pending_str(self, wnd):
        if self.pending_str:
            pending = self.pending_str
            self.pending_str = ''
            wnd.document.mode.on_str(wnd, pending)
#            if not wnd.closed:
#                wnd.update_window()
            return True

    def on_keyevent(self, wnd, event):
        if event.key == '\x1b' and event.no_trailing_char:
            return self.on_esc_pressed(wnd, event)
        else:
            return self.on_key_pressed(wnd, event)

    def _get_keybinds(self, wnd):
        return [wnd.document.mode.keybind]

    def _get_command(self, wnd, event):
        self.pending_keys.append(event.key)
        candidates = [(keys, command)
                      for keybind in self._get_keybinds(wnd)
                      for keys, command
                      in keybind.get_candidates(self.pending_keys)]

        if not candidates:
            if len(self.pending_keys) == 1:
                if isinstance(event.key, str):
                    s = event.key
                    return s, None, candidates

        elif len(candidates[0][0]) == len(self.pending_keys):
            return None, candidates[0][1], candidates

        return None, None, candidates

    def _on_str(self, wnd, s):
        for c in s:
            if c in '\t\r\n' or c >= ' ':
                self.pending_str += c

    def _keys_to_str(self, keys):
        ret = []
        alt = False
        for k in keys:
            if k == '\x1b':
                alt = True
            else:
                s = kaa.app.get_keyname(k)
                if alt:
                    s = 'alt-' + s
                    alt = False
                ret.append(s)
        return ret

    def _show_pending_keys(self, s, commands, candidates):
        if s or commands or not candidates:
            kaa.app.messagebar.set_message('')
        else:
            cur = self._keys_to_str(self.pending_keys)
            curlen = len(cur)
            nextkeys = set(self._keys_to_str(k)[curlen:curlen + 1][0]
                           for k, commands in candidates)

            msg = ' '.join(cur) + ' [%s]' % (', '.join(sorted(nextkeys)))
            kaa.app.messagebar.set_message(msg)

    def on_key_pressed(self, wnd, event):
        s, commands, candidate = self._get_command(wnd, event)
        s, commands, candidate = wnd.document.mode.on_keypressed(
            wnd, event, s, commands, candidate)

        self._show_pending_keys(s, commands, candidate)
        try:
            if s:
                self._on_str(wnd, s)
                del self.last_command_keys[:]

            elif commands:
                self.flush_pending_str(wnd)

                self.last_command_keys.append(self.pending_keys)
                del self.last_command_keys[:-3]

                wnd.document.mode.on_commands(wnd, commands)
        finally:
            if s or commands or not candidate:
                self.pending_keys = []

    def on_esc_pressed(self, wnd, event):
        kaa.app.messagebar.set_message('')
        self.pending_keys = []
        self.clear_repeat()
        del self.last_command_keys[:]

        if not wnd.closed:
            self.flush_pending_str(wnd)

        if not wnd.closed:
            # cancel mark
            wnd.screen.selection.set_mark(None)

            wnd.document.mode.on_esc_pressed(wnd, event)

    def clear_repeat(self):
        self.repeat = None
        self.repeat_str = ''
        if kaa.app.macro.is_recording():
            kaa.app.macro.record_repeatcount(None)

    def add_repeat_char(self, wnd, c):
        self.repeat_str += c
        try:
            self.set_repeat(int(self.repeat_str))
            if kaa.app.macro.is_recording():
                kaa.app.macro.record_repeatcount(self.repeat)
        except ValueError:
            self.init_repeat()

    def has_repeat(self):
        return self.repeat is not None

    def get_repeat(self):
        return self.repeat

    def set_repeat(self, n):
        self.repeat = n


class CommandMode(EditMode):
    MODENAME = 'Command'

    def on_key_pressed(self, wnd, event):
        # check if key is repeat count
        if not self.pending_keys:
            if (isinstance(event.key, str) and (event.key > '0') and
                    (event.key in string.digits)):

                self.add_repeat_char(wnd, event.key)
                return

        super().on_key_pressed(wnd, event)

    def _get_keybinds(self, wnd):
        return [wnd.document.mode.keybind_vi_commandmode,
                wnd.document.mode.keybind]

    def flush_pending_str(self, wnd):
        self.pending_str = ''


class VisualMode(CommandMode):
    MODENAME = 'Visual'

    def _get_keybinds(self, wnd):
        return [wnd.document.mode.keybind_vi_visualmode]


class VisualLinewiseMode(CommandMode):
    MODENAME = 'Visual(Line)'

    def _get_keybinds(self, wnd):
        return [wnd.document.mode.keybind_vi_visuallinewisemode]



from gappedbuf import re as gre
import collections


class Token:

    def __init__(self, name, stylename):
        self.name = name
        self.stylename = stylename

    def prepare(self, tokenizer):
        """Called to setup tokenizer"""
        self.tokenids = {}

    def re_start(self):
        """Returns regular expression to find begging of token"""

    def on_start(self, tokenizer, doc, pos, match):
        """Called when token started. Yield (pos, posto, tokenid)
        until exhausted"""

    def get_tokenids(self, highlighter, tokenizer):
        return tuple(self.tokenids.keys())

    def resume_pos(self, highlighter, tokenizer, doc, pos):
        # Returns top of current keyword
        if 0 < pos < len(doc.styles):
            p = doc.styles.rfindint(
                self.get_tokenids(highlighter, tokenizer),
                0, pos, comp_ne=True)
            if p != -1:
                return p + 1
        return 0

    def assign_tokenid(self, tokenizer, stylename):
        tokenid = tokenizer.register_tokenid(self)
        self.tokenids[tokenid] = stylename
        return tokenid

    def get_stylename(self, tokenid):
        return self.tokenids[tokenid]


class SingleToken(Token):

    def __init__(self, name, stylename, tokens):
        super().__init__(name, stylename)
        self.tokens = tokens

    def re_start(self):
        return r'({})'.format('|'.join(self.tokens))

    def prepare(self, tokenizer):
        super().prepare(tokenizer)

        self.tokenid = self.assign_tokenid(tokenizer, self.stylename)

    def on_start(self, tokenizer, doc, pos, match):
        yield (match.start(), match.end(), self.tokenid)
        return match.end(), None, False


class Keywords(SingleToken):

    """Keyword tokens:
        ex) KeyWords('keyword', 'kwdstyle', keywords=['def', 'class', 'return'])
    """

    def re_start(self):
        return r'\b({})\b'.format('|'.join(gre.escape(k) for k in self.tokens))


class Span(Token):

    def __init__(self, name, stylename, start, end, escape=None,
                 capture_end=True):

        super().__init__(name, stylename)

        self.start = start
        self.escape = escape
        if escape:
            end = '({}.)|({})'.format(gre.escape(escape), end)
        self.end = gre.compile(end, gre.X + gre.M + gre.S)
        self._capture_end = capture_end

    def prepare(self, tokenizer):
        super().prepare(tokenizer)

        self.span_start = self.assign_tokenid(tokenizer, self.stylename)
        self.span_mid = self.assign_tokenid(tokenizer, self.stylename)
        self.span_end = self.assign_tokenid(tokenizer, self.stylename)

    def re_start(self):
        return self.start

    def _is_end(self, doc, m):
        return True

    def on_start(self, tokenizer, doc, pos, match):
        yield (match.start(), match.end(), self.span_start)

        for m in self.end.finditer(doc.buf, match.end()):
            if self.escape and m.group(1) is not None:
                continue

            if match.end() != m.start():
                if not self._is_end(doc, m):
                    continue
                yield (match.end(), m.start(), self.span_mid)

            if self._capture_end:
                if m.start() != m.end():
                    yield (m.start(), m.end(), self.span_end)
                end = m.end()
            else:
                end = m.start()

            return end, None, False

        else:
            yield (match.end(), doc.endpos(), self.span_mid)
            return doc.endpos(), None, False


class SubTokenizer(Token):

    def __init__(self, name, start, tokenizer):
        super().__init__(name, '')
        self.start = start

        self.subtokenizer = tokenizer

    def prepare(self, tokenizer):
        super().prepare(tokenizer)

        self.sub_tokens = {}
        self.tokenizer = tokenizer
        self.subtokenizer.prepare(self)

    def register_tokenid(self, tokenizer, token):
        ret = self.tokenizer.highlighter.register_tokenid(self.tokenizer, self)
        self.sub_tokens[ret] = token
        return ret

    def re_start(self):
        return self.start

    def iter_subtokenizers(self, tokenizer, doc, pos, match):
        pos, tok = yield from self.subtokenizer.start(doc, pos)
        return pos, None, False

    def on_start(self, tokenizer, doc, pos, match):
        ret = yield from self.iter_subtokenizers(tokenizer, doc, pos, match)
        return ret

    def resume_pos(self, highlighter, tokenizer, doc, pos):
        # Returns top of current keyword
        if 0 < pos < len(doc.styles):
            p = doc.styles.rfindint(tuple(self.sub_tokens.keys()), 0, pos,
                                    comp_ne=True)
            if p > 0:
                return highlighter.get_resume_pos(doc, p)
        return 0

    def get_stylename(self, tokenid):
        token = self.sub_tokens[tokenid]
        if token:
            return token.get_stylename(tokenid)
        else:
            return 'default'


class SubSection(Token):

    def __init__(self, name, stylename, start, tokenizer):
        super().__init__(name, stylename)

        self.start = start
        self.tokenizer = tokenizer

    def prepare(self, tokenizer):
        super().prepare(tokenizer)
        self.section_start = self.assign_tokenid(tokenizer, self.stylename)

    def re_start(self):
        return self.start

    def on_start(self, tokenizer, doc, pos, match):
        yield (match.start(), match.end(), self.section_start)
        return match.end(), self.tokenizer, False


class EndSection(Token):

    def __init__(self, name, stylename, end):
        super().__init__(name, stylename)

        self.end = end

    def prepare(self, tokenizer):
        super().prepare(tokenizer)

        self.section_end = self.assign_tokenid(tokenizer, self.stylename)

    def re_start(self):
        return self.end

    def on_start(self, tokenizer, doc, pos, match):
        yield (match.start(), match.end(), self.section_end)
        return match.end(), None, True


class Tokenizer:
    re_starts = None

    def __init__(self, tokens, terminates=None):

        self.groupnames = {}
        self.tokens = tokens
        self.terminates = terminates

    def prepare(self, highlighter):
        self.highlighter = highlighter
        self.nulltoken = self.register_tokenid(None)

        starts = []
        for i, token in enumerate(self.tokens):
            if not token:
                continue
            token.prepare(self)
            start = token.re_start()
            if start:
                name = 'G{}'.format(i)
                self.groupnames[name] = token
                starts.append(r'(?P<{}>{})'.format(name, start))

        if self.terminates:
            starts.insert(
                0, r'(?P<{}>{})'.format('TERMINATE', self.terminates))
            self.groupnames['TERMINATE'] = None

        if starts:
            self.re_starts = gre.compile('|'.join(starts), gre.M + gre.X)

    def register_tokenid(self, obj):
        return self.highlighter.register_tokenid(self, obj)

    def start(self, doc, pos):
        if self.re_starts:
            while True:
                m = self.re_starts.search(doc.buf, pos)
                if not m:
                    break

                f, t = m.span()
                if f != pos:
                    yield (pos, f, self.nulltoken)
                    pos = f

                if m.lastgroup == 'TERMINATE':
                    return f, None

                token = self.groupnames[m.lastgroup]

                pos, childtokenizer, close = yield from token.on_start(
                    self, doc, pos, m)
                if close:
                    return pos, None

                if childtokenizer:
                    return pos, childtokenizer

        if pos != doc.endpos():
            yield (pos, doc.endpos(), self.nulltoken)
            pos = doc.endpos()

        return pos, None


class Section:
    parent = None
    end = None

    def __init__(self, pos, tokenizer):
        self.start = pos
        self.children = []
        self.tokenizer = tokenizer

    def walk(self):
        # traverse tree with depth first order
        stack = collections.deque([self])
        while stack:
            v = stack.pop()
            yield v
            stack.extend(reversed(v.children))

    def add(self, child):
        self.children.append(child)
        child.parent = self

    def get_section(self, pos):
        secs = list(self.walk())
        for sec in reversed(secs):
            if sec.start <= pos:
                if sec.end is None or pos < sec.end:
                    return sec
        return secs[0]

    def delete_after(self, pos):
        for n, c in enumerate(self.children):
            if pos < c.start:
                del self.children[n:]
                break
        section = self.get_section(pos)

        while section.parent:
            idx = section.parent.children.index(section)
            del section.parent.children[idx + 1:]
            section = section.parent


class Highlighter:

    def __init__(self, tokenizers):
        self.tokenizers = tokenizers
        self.max_tokenid = 1
        self.tokenids = {}

        for tokenizer in self.tokenizers:
            tokenizer.prepare(self)

        self.section = None
        self.updatepos = 0
        self._highlighter = None

    def close(self):
        self._highlighter = self.section = self.tokenizer = None

    def register_tokenid(self, tokenizer, token):
        tokenid = self.max_tokenid
        self.max_tokenid += 1

        self.tokenids[tokenid] = (tokenizer, token)
        return tokenid

    def get_token(self, tokenid):
        if tokenid in self.tokenids:
            return self.tokenids[tokenid][1]

    def update_style(self, doc, batch=None):
        # todo: store 'doc' as attribute
        if not self._highlighter:
            self._highlighter = self.highlight(doc, self.updatepos)

        updatefrom = doc.endpos()
        updateto = 0
        updated = False
        try:
            for n, (start, end, style) in enumerate(self._highlighter):
                updated = True
                updatefrom = min(start, updatefrom)
                updateto = max(end, updateto)
                doc.styles.setints(start, end, style)
                if batch and (n > batch):
                    return True
            else:
                return False

        finally:
            if doc.endpos() == 0 or updated and (updatefrom != updateto):
                doc.style_updated(updatefrom, updateto)
                self.updatepos = updateto

    def highlight(self, doc, pos):
        if not self.tokenizers:
            return

        if not self.section:
            pos = 0
            section = Section(0, self.tokenizers[0])
            self.section = section
        else:
            self.section.delete_after(pos)
            section = self.section.get_section(pos)

        section.end = None
        while section:
            pos, childtokenizer = yield from section.tokenizer.start(doc, pos)
            if childtokenizer:
                childsection = Section(pos, childtokenizer)
                section.add(childsection)
                section = childsection
            else:
                section.end = pos
                section = section.parent

    def get_resume_pos(self, doc, pos):
        if pos == 0:
            return pos

        # check a character proceeding to updated pos
        pos -= 1
        style = doc.styles.getints(pos, pos + 1)[0]
        if style == 0:
            # not highlighted yet.
            p = doc.styles.rfindint([0], 0, pos, comp_ne=True)
            if p != -1:
                return p + 1
            return 0

        pair = self.tokenids.get(style)
        if not pair:
            # Invalid. this style is not set by tokenizer.
            return 0

        tokenizer, token = self.tokenids.get(style)
        if not token:
            p = doc.styles.rfindint([style], 0, pos, comp_ne=True)
            if p != -1:
                return p + 1
            return 0

        return token.resume_pos(self, tokenizer, doc, pos)

    def updated(self, doc, pos, inslen, dellen):
        if pos <= self.updatepos:
            self.updatepos = self.get_resume_pos(doc, pos)
            self._highlighter = None



import os
import kaa
from kaa import document
from kaa.theme import Theme, Style
from kaa.ui.dialog import dialogmode
from kaa.filetype.default import keybind
from kaa.ui.itemlist import itemlistmode

FileInfoThemes = {
    'basic': [
        Style('button', 'Base3', 'Blue'),
        Style('shortcut', 'Base3', 'Orange', bold=True,
              underline=True, nowrap=True),
    ],
}


class FileInfoMode(dialogmode.DialogMode):
    autoshrink = True

    KEY_BINDS = [
        keybind.cursor_keys,
        keybind.edit_command_keys,
        keybind.emacs_keys,
        keybind.macro_command_keys,
    ]

    def __init__(self, target):
        super().__init__()

        self.target = target

    def close(self):
        super().close()
        self.target = None
        kaa.app.messagebar.set_message("")

    def init_themes(self):
        super().init_themes()
        self.themes.append(FileInfoThemes)

    def build_doc(self):
        self.document.delete(0, self.document.endpos())

        filename = self.target.document.get_filename()
        dir, fname = os.path.split(filename)

        mode = self.target.document.mode
        fileinfo = self.target.document.fileinfo

        with dialogmode.FormBuilder(self.document) as f:
            f.append_text('caption', '    File name: ')
            f.append_text('default', ' {}\n'.format(fname.replace('&', '&&')))

            f.append_text('caption', '    Directory: ')
            f.append_text('default', ' {}\n'.format(dir.replace('&', '&&')))

            f.append_text('default', '\n')

            f.append_text('caption', '         Mode: ')
            f.append_text('default', ' {}     '.format(mode.MODENAME))
            f.append_text('button', '[Change &Mode]\n',
                          on_shortcut=self._on_update_mode,
                          shortcut_style='shortcut',
                          shortcut_need_alt=False)

            f.append_text('caption', '     Encoding: ')
            f.append_text('default', ' {}\n'.format(
                fileinfo.encoding if fileinfo else ''))

            f.append_text('caption', '      Newline: ')
            f.append_text('default', ' {}\n'.format(
                fileinfo.newline if fileinfo else ''))

            f.append_text('caption', '  Line number: ')
            f.append_text('default', ' {}       '.format(
                'Show' if mode.SHOW_LINENO else 'Hide'))
            f.append_text('button', '[Change &Line number]\n',
                          on_shortcut=self._on_update_lineno,
                          shortcut_style='shortcut',
                          shortcut_need_alt=False)

            f.append_text('default', '\n')

            f.append_text('caption', '    Tab width: ',
                          shortcut_style='shortcut')
            f.append_text('default', ' {}          '.format(mode.tab_width))
            f.append_text('button', '[Change &Tab width]\n',
                          on_shortcut=self._on_update_tab,
                          shortcut_style='shortcut',
                          shortcut_need_alt=False)

            f.append_text('caption', ' Indent width: ',
                          shortcut_style='shortcut')
            f.append_text('default', ' {}          '.format(mode.indent_width))
            f.append_text('button', '[Change &Indent width]\n',
                          on_shortcut=self._on_update_indent,
                          shortcut_style='shortcut',
                          shortcut_need_alt=False)

            f.append_text('caption', '      Use tab: ',
                          shortcut_style='shortcut')
            f.append_text('default', ' {}        {}'.format(
                'Yes' if mode.indent_tab else 'No', '' if mode.indent_tab else ' '))
            f.append_text('button', '[Change &Use tab]\n',
                          on_shortcut=self._on_update_usetab,
                          shortcut_style='shortcut',
                          shortcut_need_alt=False)

            f.append_text('caption', '  Auto Indent: ',
                          shortcut_style='shortcut')
            f.append_text('default', ' {}        {}'.format(
                'Yes' if mode.auto_indent else 'No', '' if mode.auto_indent else ' '))
            f.append_text('button', '[&Auto Indent]\n',
                          on_shortcut=self._on_update_autoindent,
                          shortcut_style='shortcut',
                          shortcut_need_alt=False)

    def is_cursor_visible(self):
        return 0   # hide cursor

    def on_str(self, wnd, s):
        # does nothing
        pass

    def on_esc_pressed(self, wnd, event):
        super().on_esc_pressed(wnd, event)
        popup = wnd.get_label('popup')
        popup.destroy()

    def _on_update_mode(self, c):
        modes = []
        for pkg in kaa.app.config.get_mode_packages():
            modes.append(pkg.FileTypeInfo.get_modetype())

        names = [mode.MODENAME for mode in modes]

        def callback(n):
            if n is None:
                return

            mode = modes[n]
            if mode is not type(self.target.document.mode):
                self.target.document.setmode(mode())
                self.build_doc()

        doc = itemlistmode.ItemListMode.build(
            'Select file mode:',
            names,
            names.index(self.target.document.mode.MODENAME),
            callback)

        kaa.app.show_dialog(doc)

    def _on_update_lineno(self, c):
        values = ['Hide', 'Show']

        def callback(n):
            if n is None:
                return

            v = bool(n)
            if self.target.document.mode.SHOW_LINENO != v:
                self.target.document.mode.SHOW_LINENO = v
                self.target.document.update_screen(0, 0, 0)
                self.build_doc()

        doc = itemlistmode.ItemListMode.build(
            'Use tab for indent?',
            values,
            int(self.target.document.mode.SHOW_LINENO),
            callback)

        kaa.app.show_dialog(doc)

    def _on_update_tab(self, c):
        values = [str(i) for i in range(2, 9)]

        def callback(n):
            if n is None:
                return

            width = int(values[n])
            if self.target.document.mode.tab_width != width:
                self.target.document.mode.tab_width = width
                self.target.document.update_screen(0, 0, 0)
                self.build_doc()

        doc = itemlistmode.ItemListMode.build(
            'Select tab width:',
            values,
            values.index(str(self.target.document.mode.tab_width)),
            callback)

        kaa.app.show_dialog(doc)

    def _on_update_indent(self, c):
        values = [str(i) for i in range(2, 9)]

        def callback(n):
            if n is None:
                return

            width = int(values[n])
            if self.target.document.mode.indent_width != width:
                self.target.document.mode.indent_width = width
                self.target.document.update_screen(0, 0, 0)
                self.build_doc()

        doc = itemlistmode.ItemListMode.build(
            'Select indent width:',
            values,
            values.index(str(self.target.document.mode.indent_width)),
            callback)

        kaa.app.show_dialog(doc)

    def _on_update_usetab(self, c):
        values = ['No', 'Yes']

        def callback(n):
            if n is None:
                return

            v = bool(n)
            if self.target.document.mode.indent_tab != v:
                self.target.document.mode.indent_tab = v
                self.target.document.update_screen(0, 0, 0)
                self.build_doc()

        doc = itemlistmode.ItemListMode.build(
            'Use tab for indent?',
            values,
            int(self.target.document.mode.indent_tab),
            callback)

        kaa.app.show_dialog(doc)

    def _on_update_autoindent(self, c):
        values = ['No', 'Yes']

        def callback(n):
            if n is None:
                return

            v = bool(n)
            if self.target.document.mode.auto_indent != v:
                self.target.document.mode.auto_indent = v
                self.target.document.update_screen(0, 0, 0)
                self.build_doc()

        doc = itemlistmode.ItemListMode.build(
            'Use auto indentation?',
            values,
            int(self.target.document.mode.auto_indent),
            callback)

        kaa.app.show_dialog(doc)


def show_fileinfo(target):
    buf = document.Buffer()
    doc = document.Document(buf)
    mode = FileInfoMode(target)
    doc.setmode(mode)

    mode.build_doc()
    kaa.app.show_dialog(doc)
    kaa.app.messagebar.set_message(
        'Hit underlined character to update.')



import os
from kaa import config


class TestHistory:
    def test_history(self):
        storage = config.KaaHistoryStorage('')
        try:
            hist = storage.get_history('hist1')
            hist.add('1', 1)
            hist.add('2', 2)
            hist.add('1', 1)

            assert hist.get() == [('1', 1), ('2', 2)]
            assert hist.find('1') == 1
            
            storage.flush()
            assert hist.get() == [('1', 1), ('2', 2)]
            assert hist.find('1') == 1

            hist.add('1', 1)
            hist.add('3', 3)
            assert hist.get() == [('3', 3), ('1', 1), ('2', 2),]

            storage.flush()
            assert hist.get() == [('3', 3), ('1', 1), ('2', 2),]
            
        finally:
            storage.close()

    def test_histclose(self):
        storage = config.KaaHistoryStorage('test.db')
        try:
            hist = storage.get_history('hist2')

            for i in range(config.History.MAX_HISTORY * 3):
                hist.add(str(i))
        finally:
            storage.flush()
            storage.close()

        storage = config.KaaHistoryStorage('test.db')
        try:
            hist = storage.get_history('hist2')

            assert len(hist.get()) == config.History.MAX_HISTORY
            hist.close()
        finally:
            storage.close()
            os.unlink('test.db')



from kaa.filetype import filetypedef

class FileTypeInfo(filetypedef.FileTypeInfo):
    FILE_EXT = {'.css'}

    @classmethod
    def get_modetype(cls):
        from kaa.filetype.css.cssmode import CSSMode
        return CSSMode
